#!/usr/local/bin/ python3
#This program is a python script to create a data base of real observations at different resolutions.
# It take 6 publicly available well resolved galaxy data cubes and smooths them to 3,4,5,6,7,8,10,12,16 beams across the major axis based on the extend of the best fit model.
# The galaxies used are

from pyHIARD.constants import c_kms, H_0, G_agc
from pyHIARD import Templates as templates
from multiprocessing import Pool, get_context
from astropy.wcs import WCS
from astropy.io import fits
import copy
import importlib
import numpy as np
import os
import psutil
import pyHIARD.common_functions as cf
import scipy.ndimage
import sys
import warnings

# be explicit
try:
    import importlib.resources as import_res
except ImportError:
    import importlib_resources as import_res

class ProgramError(Exception):
    pass


def add_template(cfg, path_to_resources, existing_galaxies):
    confirm_directory = False
    while not confirm_directory:
        confirm_directory = cf.get_bool(
            f"We are working in the directory {cfg.general.main_directory}. Is this correct? (Yes/No, default =True)", default=True)
        if not confirm_directory:
            cfg.general.main_directory = input(
                f"Which directory would you like to work in?")
        else:
            while not os.path.isdir(cfg.general.main_directory):
                print(
                    f'The directory {cfg.general.main_directory} does not exist please provide the correct directory')
                cfg.general.main_directory = input(
                    "Please provide the directory where to create the database :")
            confirm_directory = True

    galaxy_parameters = {'Galaxy': None, 'DHIkpc': None,
        'Distance': None, 'Original_Model': None, 'RMS': None, 'MHI':  None}
    galaxy_translation = {'Galaxy': 'the name to use in the package',
                          'DHIkpc':  'the HI diameter in kpc',
                          'Distance': 'distance in Mpc',
                          'Original_Model': 'the format of the orginal model (Tirific, Barolo, Rotcur)',
                          'RMS': 'the noise in the cube in data units',
                          'MHI':  'the HI mass'}

    existing_galaxies_low = [x.lower() for x in existing_galaxies]

    for key in galaxy_parameters:
        galaxy_parameters[key] = input(
            f"Please provide {galaxy_translation[key]} of the galaxy:")
        if key == 'Galaxy':
            if galaxy_parameters[key].lower() in existing_galaxies_low:
                remove_galaxy = cf.get_bool(
                    f"The galaxy {galaxy_parameters[key]} already exists, do you want to delete it? (default = no)", default=False)
                if remove_galaxy:
                    removed = remove_template(
                        galaxy_parameters[key], path_to_resources, existing_galaxies)
                    if not removed:
                        while galaxy_parameters[key].lower() in existing_galaxies_low:
                            galaxy_parameters[key] = input(
                                f"Please provide a different name (current = {galaxy_parameters[key]}).")

                else:
                    while galaxy_parameters[key].lower() in existing_galaxies_low:
                        galaxy_parameters[key] = input(
                            f"Please provide a different name (current = {galaxy_parameters[key]}).")

        if key == 'Original_Model':
            acceptable_model = False
            while not acceptable_model:
                if galaxy_parameters[key].lower() == 'tirific':
                    galaxy_parameters[key] = 'Tir'
                    acceptable_model = True
                elif galaxy_parameters[key].lower() == 'rotcur':
                    galaxy_parameters[key] = 'RC'
                    acceptable_model = True
                elif galaxy_parameters[key].lower() == 'barolo':
                    galaxy_parameters[key] = 'Bar'
                    acceptable_model = True
                else:
                    galaxy_parameters[key] = input(
                        f"{galaxy_parameters[key]} is not yet a model pyHIARD can process please type TiRiFiC, Barolo or Rotcur as an input model: ")

            model_file = input(f'Please provide the model text file:')
            while not os.path.isfile(f"{cfg.general.main_directory}{model_file}"):
                model_file = input(
                    f'Print we can not find the file {cfg.general.main_directory}{model_file}, please provide the correct name with a path from {cfg.general.main_directory}:')

            #try:
            test = cf.load_text_model(
            f"{cfg.general.main_directory}{model_file}", package_file=False,
             type=galaxy_parameters[key], Variables=['RADI', 'VROT', 'PA', 'INCL', 'XPOS', 'YPOS',
             'VSYS', 'VROT_2', 'PA_2', 'INCL_2', 'XPOS_2', 'YPOS_2', 'VSYS_2', 'Z0', 'SDIS', 'Z0_2', 'SDIS_2', 'CONDISP', 'SBR', 'SBR_2'])
            #except:
            #    print(f"We cannot read your file, please provide a standard TiRiFiC, Barolo or RotCur model.")
            #    print("We are exiting pyHIARD")
            #    sys.exit()

    input_fits_file = input("Please provide the galaxy fits file:")
    while not os.path.isfile(f"{cfg.general.main_directory}{input_fits_file}"):
        input_fits_file = input(
            f"We cannot find {cfg.general.main_directory}{input_fits_file}. Please provide the correct path from {cfg.general.main_directory}.")

    try:
        Cube = fits.open(f"{cfg.general.main_directory}{input_fits_file}",
                         uint=False, do_not_scale_image_data=True, ignore_blank=True)
        length = Cube[0].header['NAXIS3']
    except:
        print(f"We cannot read your fits file, please provide a standard fits file with 3 axes.")
        print("We are exiting pyHIARD")
        sys.exit()

    #galaxy_parameters = {'Galaxy': 'New_Galaxy', 'DHIkpc': '9.6', 'Distance': '4.1', 'Original_Model': 'Tir', 'RMS': '0.00038', 'MHI': '0.54e9'}
    #read or template and modify it
    '''    
    if sys.version_info[0][:3]) < 3.9:
        with import_res.open_text(templates, 'roc_galaxy_template.py') as tmp:
            module_template = tmp.readlines()
    else:
    '''    
    with import_res.files(templates).joinpath('roc_galaxy_template.py').open('r') as tmp:
        module_template = tmp.readlines()

    galaxy_line = "galaxy_parameters = {"
    for key in galaxy_parameters:
        if key in ['Galaxy', 'Original_Model']:
            galaxy_line = f"{galaxy_line}'{key}': '{galaxy_parameters[key]}', "
        else:
            galaxy_line = f"{galaxy_line}'{key}': {galaxy_parameters[key]}, "
    galaxy_line = galaxy_line[:-2]+"}"

    for i, line in enumerate(module_template):
        start = line.split('=')
        if start[0].strip() == 'galaxy_parameters':
            module_template[i] = galaxy_line
        if 'Input_Name' in line:
            line = line.replace('Input_Name', galaxy_parameters['Galaxy'])
            module_template[i] = line

    #Create the new directory and place all files in there
    new_resource = os.path.join(
        path_to_resources, galaxy_parameters['Galaxy'])+'/'
    cf.create_directory(galaxy_parameters['Galaxy'], path_to_resources)

    fits.writeto(
        f'{new_resource}/{galaxy_parameters["Galaxy"]}_Original.fits', Cube[0].data, Cube[0].header, overwrite=False)
    Cube.close()
    #Mask_Inner, Mask_Outer = cf.create_masks(
    #    new_resource, cfg.general.main_directory, galaxy_parameters['Galaxy'], sofia_call=cfg.general.sofia2)
    ext = {'Tir': 'def', 'Bar': 'txt', 'RC': 'rotcur'}
    os.system(
        f"cp {cfg.general.main_directory}{model_file} {new_resource}{galaxy_parameters['Galaxy']}.{ext[galaxy_parameters['Original_Model']]}")
    with open(f"{new_resource}{galaxy_parameters['Galaxy']}.py", 'w') as f:
        f.writelines(module_template)
    with open(f"{new_resource}__init__.py", 'w') as f:
        f.write(f'   ')

add_template.__doc__ = f'''
NAME:
    add_template

PURPOSE:
    add a galaxy template to the package

CATEGORY:
   roc

INPUTS:
    cfg=  input configuration
    path_to_resources = path to the resources cube directory
    existing_galaxies = list of existing templates

OPTIONAL INPUTS:

OUTPUTS:
    template is added inside the package

OPTIONAL OUTPUTS:

PROCEDURES CALLED:
   Unspecified

NOTE:
'''


def beam_templates(beam_req, Galaxy_Template_In, max_degradation_factor,\
                        main_directory):
    Galaxy_Template = copy.deepcopy(Galaxy_Template_In)
    # first we need to calculate the shift to  apply
    print(f"We are processing the template {Galaxy_Template['Name']} with {beam_req} beams across the major axis.")
    #First we calculate the degradation we want

    new_beam = [(Galaxy_Template['Original_DHI_arcsec'])
                 / (3600.*beam_req)]  # let's do beams in degrees
    initial_factor = new_beam[0] / \
        Galaxy_Template['Galaxy_Template_Header']['BMAJ']
    # if this is larger than the maximal degradation we smooth and regird to safe time and memory
    #fits.writeto('Before_Regrid.fits', Galaxy_Template['Galaxy_Template_Cube'],
    #             Galaxy_Template['Galaxy_Template_Header'], overwrite=True)

    #print(
    #    f"Currently the template is {Galaxy_Template['Galaxy_Template_Cube'].size*Galaxy_Template['Galaxy_Template_Cube'].itemsize/(1024**2)} Mb")
    if initial_factor > max_degradation_factor:
        smooth_factor=initial_factor/max_degradation_factor
        extended_template_cube, extended_template_header, pixel_buffer=\
            extend_cube(Galaxy_Template['Galaxy_Template_Cube'],
                                 Galaxy_Template['Galaxy_Template_Header'],
                                 new_beam = Galaxy_Template['Galaxy_Template_Header']['BMAJ']
                                  * smooth_factor)

        template_cube, template_hdr, template_noise=smooth_and_regrid(extended_template_cube,
            extended_template_header, regrid=True, track_noise = Galaxy_Template['Original_Noise'],
            factor=smooth_factor)
        del extended_template_cube
        del extended_template_header

        del Galaxy_Template['Galaxy_Template_Cube']
        del Galaxy_Template['Galaxy_Template_Header']

        regrid_factor = new_beam[0]/(template_hdr['BMAJ'])
    else:
        #if we don't regrid the factors are the same
        smooth_factor = 1.
        regrid_factor = initial_factor
        template_cube = Galaxy_Template['Galaxy_Template_Cube']
        template_hdr = Galaxy_Template['Galaxy_Template_Header']
        template_noise = Galaxy_Template['Original_Noise']

    template_mask = cf.get_mask(template_cube,factor = initial_factor)

    new_mean = cf.get_mean_flux(template_cube,Mask=template_mask)

    # And the new bmin has the same factor
    new_beam.append(template_hdr['BMIN'] * regrid_factor)
    # calclate the pixels in the new beam area
    new_pixels_per_beam=cf.get_beam_area_in_pixels(
        template_hdr, beam = np.array(new_beam, dtype=float))


    #Start shifting the template
    # first a new distance
    new_distance=Galaxy_Template['Original_Distance']*initial_factor
    #And the corresponding systemic on the Hubble Flow +the original peculiar velocity
    peculiar_velocity=Galaxy_Template['Original_Vsys'] - \
        H_0*Galaxy_Template['Original_Distance']
    new_systemic=H_0*new_distance+peculiar_velocity
    # And the new redshift (This is actually 1+z)
    new_z=np.sqrt((1. + new_systemic / c_kms) / (1. - new_systemic / c_kms))
    #print(f"Is the z reasonable {new_z}")
    #And the shifted flux value adjusting the flux and the dimming factor
    #new_total_flux = Galaxy_Template['Original_Total_Flux_In'] / \
    #    (initial_factor**2)
    #
    new_DHI_arcsec=cf.convertskyangle(cf.convertskyangle(Galaxy_Template['Original_DHI_arcsec'],\
                    distance=Galaxy_Template['Original_Distance']),\
                    distance = new_distance, physical = True)
    #And now we want to shift the actual cube Which simply mean reducing the flux and account for Tollman dimming through the z factor
    new_template_cube=copy.deepcopy(template_cube)/(initial_factor**2) *\
            ((Galaxy_Template['Original_z']**4)/(new_z**4))
    #Calculate the noise corresponding to this

    new_noise=float(template_noise) /\
                (initial_factor**2) *\
                ((Galaxy_Template['Original_z']**4)/(new_z**4))
    new_mean=float(new_mean) /\
            (initial_factor**2) *\
            ((Galaxy_Template['Original_z']**4)/(new_z**4))

    #And update the header to the corresponding values

    template_hdr["CRVAL3"]=new_systemic
    #Now we update the TRM to new values
    new_TRM_model=copy.deepcopy(Galaxy_Template['Galaxy_TRM_Model'])
    for i, beam in enumerate(['BMAJ', 'BMIN']):
        new_TRM_model[beam]=f'{beam}= {new_beam[i]/initial_factor*3600.:.2f}'
    new_TRM_model['DISTANCE']=f'DISTANCE = {new_distance:.2f}'
    #And the radi, they should be stored in kpc in the uniform template
    radius=[float(x) for x in new_TRM_model['RADI'].split('=')[1].split()]
    conv_radi=cf.convertskyangle(
        radius, distance = new_distance, physical = True)
    new_TRM_model['RADI']='RADI = '+' '.join([str(e) for e in conv_radi])
    #same for the scale heights
    for i, var in enumerate(['', '_2']):
        scaleheight=[float(x)
                             for x in new_TRM_model[f'Z0{var}'].split('=')[1].split()]
        con_hz= cf.convertskyangle(
            scaleheight, distance=new_distance, physical=True)
        if np.sum(scaleheight) != 0.:
            new_TRM_model[f'Z0{var}'] = f'Z0{var} = {" ".join(str(e) for e in con_hz if e != 0.)}'
        else:
            new_TRM_model[f'Z0{var}'] = f'Z0{var} = 0.'
        new_TRM_model[f'VSYS{var}'] = f'VSYS{var} = {new_systemic}'
        sbr = [float(x)
                     for x in new_TRM_model[f'SBR{var}'].split('=')[1].split()]
        if np.sum(sbr) != 0.:
            new_TRM_model[f'SBR{var}'] =\
             f'SBR{var} = {" ".join(str(e*((Galaxy_Template["Original_z"]**4)/(new_z**4))) for e in sbr if e != 0.)}'
        else:
            new_TRM_model[f'SBR{var}'] = f'SBR{var} = 0.'
    # And we need to adjust the header for our new cube
    # We want to make a template at the final resolution to get some values from
    # Frist we ensure some pixel buffer after smoothing
    extended_new_template_cube, extended_new_template_hdr,pixel_buffer ,extended_mask= extend_cube(
            new_template_cube, template_hdr, new_beam=new_beam[0],Mask=template_mask,rotation_pa=template_hdr['BPA'])
    #print(f"We extended these image with {pixel_buffer} current size {extended_new_template_cube.shape} before {new_template_cube.shape}")
    template_at_final_res, hdr_at_final_res,noise_at_final_res = smooth_and_regrid(extended_new_template_cube,
        extended_new_template_hdr, factor=regrid_factor, regrid = False,track_noise=new_noise)

    # Let's make a mask from this smoothed cube to calculated the things we achieve
    Final_Mask=copy.deepcopy(template_at_final_res)
    Final_Mask[template_at_final_res > np.mean(template_at_final_res[template_at_final_res > 0.])/2.]=1
    Final_Mask[template_at_final_res < np.mean(
        template_at_final_res[template_at_final_res > 0.]) / 2.]=0
    Final_Mean=cf.get_mean_flux(template_at_final_res, Mask = Final_Mask)
    Final_Mask = np.array(Final_Mask,dtype=int)
    pxbp = cf.get_beam_area_in_pixels(hdr_at_final_res)
    Final_Total_Flux = np.sum(template_at_final_res)/pxbp *hdr_at_final_res['CDELT3']
    #Now we can check if this galaxy already exists
    if -1. in Galaxy_Template['Requested_SNR']:
        reqnoise = float(Final_Mean/noise_at_final_res)
        dirstring = f"{Galaxy_Template['Name']}_{beam_req:.1f}Beams_{reqnoise:.1f}SNR"
        galaxy_dir =f"{main_directory}{dirstring}/"
        galaxy_dir_exists = os.path.isdir(galaxy_dir)
        if galaxy_dir_exists:
    # Do we have a cube
            galaxy_cube_exist = os.path.isfile(f"{galaxy_dir}Convolved_Cube_Gauss.fits")
            if galaxy_cube_exist:
                Galaxy_Template['Requested_SNR'].remove(-1.)
                print(f"The galaxy {dirstring} galaxy appears fully produced")
    if len(Galaxy_Template['Requested_SNR']) == 0.:
        print(f"All galaxies for template {Galaxy_Template['Name']} with {beam_req} beams across the major axis are already produced.")
        return "EMPTY"


    #These values should be corrected for the increased beam
    del template_at_final_res
    del hdr_at_final_res
    #fits.writeto('Test_New_Shift.fits',extended_new_template_cube,extended_new_template_hdr,overwrite=True)
    Template_Dictionary={'Name': Galaxy_Template['Name'], 'Beams': beam_req, 'Shifted_Template_Cube': extended_new_template_cube,\
                           'Shifted_Template_Header': extended_new_template_hdr, 'Final_Mask': Final_Mask,\
                           'Pixel_Buffer': pixel_buffer,'Final_DHI_arcsec': new_DHI_arcsec,\
                           'Shifted_TRM_Model': new_TRM_model, 'Shifted_Mask': extended_mask,\
                           'M_HI': Galaxy_Template['M_HI'],'Final_Noise': noise_at_final_res,\
                           'Shift_Factor': [initial_factor, regrid_factor],\
                           'New_Beam': [*new_beam, extended_new_template_hdr["BPA"]],\
                           'Shifted_Original_Noise': new_noise, 'Final_Mean_Flux': Final_Mean,\
                           'Final_Distance': new_distance,\
                           'Requested_SNR': Galaxy_Template['Requested_SNR'], 'Disclaimer': Galaxy_Template['Disclaimer']
                          }

    del Galaxy_Template
    return Template_Dictionary
beam_templates.__doc__ = f'''
NAME:
    beam_templates(cfg,beams,Galaxy_Template)

PURPOSE:
    construct all differnt beam templates

CATEGORY:
   roc

INPUTS:
    cfg=  input configuration
    path_to_resources = path to the resources cube directory
    existing_galaxies = list of existing templates

OPTIONAL INPUTS:

OUTPUTS:
    template is added inside the package

OPTIONAL OUTPUTS:

PROCEDURES CALLED:
   Unspecified

NOTE:
'''

def calculate_processes(templates,beams,SNR):
    available_memory = psutil.virtual_memory().total/2**30
    max_no_process = int(np.floor(available_memory/3.5)) # A single process run peaked at 7 Gb so this should suffice
    no_SNR_process = len(SNR)
    no_beam_process = 1
    no_template_process = 1
    if no_SNR_process > (max_no_process-2):
        no_SNR_process = max_no_process-2
    if (max_no_process-1) > 2.*no_SNR_process and no_SNR_process > 1:
        if len(beams)*no_SNR_process > max_no_process:
            left_process =  max_no_process-no_SNR_process-1
            while left_process > no_SNR_process:
                no_beam_process +=1
                left_process -= no_SNR_process
        else:
            no_beam_process = len(beams)
    if max_no_process > 2.*no_SNR_process*no_beam_process and no_beam_process > 1:
        if len(templates)*no_beam_process*no_SNR_process > max_no_process:
            left_process =  max_no_process-(no_SNR_process*no_beam_process)
            while left_process > no_SNR_process*no_beam_process:
                no_template_process +=1
                left_process -= no_SNR_process*no_beam_process
        else:
            no_template_process = len(templates)
    return no_template_process,no_beam_process,no_SNR_process


calculate_processes.__doc__ = f'''
NAME:
    calculate_processes(templates,beams,SNR)

PURPOSE:
    calculate the number of processes in each pool and limit the total to not go beyond the maximum memory

CATEGORY:
   roc

INPUTS:
    templates = templates to process
    beams = beams for each template
    SNR = all SNRs to process

OPTIONAL INPUTS:

OUTPUTS:
    template is added inside the package

OPTIONAL OUTPUTS:

PROCEDURES CALLED:
   Unspecified

NOTE:
'''



def check_templates(name,path_to_resources,work_dir,sofia_call='sofia2'):

    #First check if the main cube is there
    file_exists = os.path.isfile(f"{path_to_resources}{name}/{name}.fits")
    file_exists_uni = os.path.isfile(f"{path_to_resources}{name}/{name}_uniform.def")
    if not file_exists or not file_exists_uni:
        galaxy_module = importlib.import_module(f'pyHIARD.Resources.Cubes.{name}.{name}')
        Template_All=galaxy_module.get_data(work_dir,sofia_call)
        if not file_exists_uni:
            Model_Template = get_main_template(name,Template_All[0].header,galaxy_module)
        Template_All.close()

    return file_exists
check_templates.__doc__= f'''
NAME:
   check_templates(name,path_to_resources)

PURPOSE:
    Check that all the templates and their masks exist, otherwise download and create them now

CATEGORY:
   roc

INPUTS:
    name = basename of template
    path_to_resources = path to the cubes directory

OPTIONAL INPUTS:

OUTPUTS:
    boolean to indicate succes

OPTIONAL OUTPUTS:

PROCEDURES CALLED:
   Unspecified

NOTE:
'''
def create_final_cube(required_noise,main_directory,Galaxy_Template_In,font_file):
    Galaxy_Template = copy.deepcopy(Galaxy_Template_In)
    # Make a new directory
    native_noise = False
    if required_noise == -1.:
        native_noise = True
        required_noise = Galaxy_Template['Final_Mean_Flux']/Galaxy_Template['Final_Noise']

    dirstring = f"{Galaxy_Template['Name']}_{Galaxy_Template['Beams']:.1f}Beams_{required_noise:.1f}SNR"
    print(f'''{'Processing the template':>25s} = {Galaxy_Template['Name']}
{'Beams':>25s} = {Galaxy_Template['Beams']}
{'SNR':>25s} = {required_noise}''')
    galaxy_dir =f"{main_directory}{dirstring}/"
    galaxy_dir_exists = os.path.isdir(galaxy_dir)
    if not galaxy_dir_exists:
        os.system(f"mkdir {galaxy_dir}")
    #Place the disclaimer about the cube
    Galaxy_Template['Disclaimer'](galaxy_dir)
    # The noise in the final cube should be
    final_req_noise = Galaxy_Template['Final_Mean_Flux']/required_noise
    # We first find the noise in pixels that matches this
    #As we will be adding to the uncorrected cube we need to convert the noise back to its uncorrected value.
    # first we estimate what will be the noise we want for a cube with the resolution of the template
    # so we scale with the beam area new_beam/old_beam
    shifted_template_pix_beam = cf.get_beam_area_in_pixels(Galaxy_Template['Shifted_Template_Header'])
    new_beam_pix_beam = cf.get_beam_area_in_pixels(Galaxy_Template['Shifted_Template_Header'], beam=Galaxy_Template['New_Beam'][0:2])
    shifted_req_noise = final_req_noise *shifted_template_pix_beam/new_beam_pix_beam

    # The pixelated noise estimated is
    #First the smoothing in pixels required to get to our new beam
    new_pixel_sigma=(np.array(Galaxy_Template['New_Beam'][0:1],dtype=float)/ \
        np.abs(np.array([Galaxy_Template['Shifted_Template_Header'][key] for key in ['CDELT2','CDELT1'] ],dtype=float)))/\
        (2. * np.sqrt(2. * np.log(2.)))
    #And the sigma to get to the shifted template resolution
    template_pixel_sigma=(np.array([Galaxy_Template['Shifted_Template_Header'][key] for key in ['BMAJ','BMIN'] ],dtype=float)/ \
        np.abs(np.array([Galaxy_Template['Shifted_Template_Header'][key] for key in ['CDELT2','CDELT1'] ],dtype=float)))/\
        (2. * np.sqrt(2. * np.log(2.)))
    # Fine tune the pixel noise estimate to achieve the final noise
    #setup the random number generator
    achieved_shifted_noise,pixel_noise = cf.calculate_pixel_noise(shifted_req_noise,template_pixel_sigma)

    # If this noise differs significantly from the original noise in the cube then we want to add noise to the emission part as well.
    diff_noise = achieved_shifted_noise - Galaxy_Template['Shifted_Original_Noise']
    # If this difference is less than 10 % we will ignore it
    if abs(diff_noise) < achieved_shifted_noise/10. or native_noise:
        diff_noise = 0.
    #print(f'''We will get the noise at the shifted template resolution to  {achieved_shifted_noise} Jy/beam.
#This is a difference of {diff_noise} Jy/beam with the noise in the template at this shift.''')
    # If the new noise is smaller than the input noise we get into trouble an we do not want to return ProgramError(f"This should not happen") as there would be a higher noise on the emission
    if diff_noise < 0:
        with open(f"{galaxy_dir}Why_This_Galaxy_Is_Not_There.txt",'w') as file:
            file.write(f'''Your requested noise is lower than the input noise hence the emission would be too noisy please lower SNR.
The requested noise is {achieved_shifted_noise} and the original noise translated to this shift is {Galaxy_Template['Shifted_Original_Noise']}.
We continue with the next SNR value.''')
        return "EMPTY"
    #making a random number generator
    rng = np.random.default_rng()
    # calculate the noise to be added to the template
    # If we want to add noise  we construct a new noise cube for this purpose
    if diff_noise > 0.:
        achieved_diff_noise,pixel_diff_noise = cf.calculate_pixel_noise(diff_noise,template_pixel_sigma)
        #create a cube with this noise
        diff_noise_cube = rng.normal(scale=pixel_diff_noise, size=Galaxy_Template['Shifted_Template_Cube'].shape)
        # smooth it to the template resolution
        diff_noise_template = scipy.ndimage.gaussian_filter(diff_noise_cube,\
                        sigma=(0,*template_pixel_sigma),order=0)
        #We add this to the emission
        Galaxy_Template['Shifted_Template_Cube'][Galaxy_Template['Shifted_Template_Cube'] != 0.] = \
            Galaxy_Template['Shifted_Template_Cube'][Galaxy_Template['Shifted_Template_Cube'] != 0.] +\
            diff_noise_template[Galaxy_Template['Shifted_Template_Cube'] != 0.]
        del diff_noise_cube
        del diff_noise_template

    # create an extended cube with pixelize noise at the template resolution
    noise_template =  rng.normal(scale=pixel_noise, size=Galaxy_Template['Shifted_Template_Cube'].shape)
    #smooth to the template resolution
    noise_template = scipy.ndimage.gaussian_filter(noise_template,\
                        sigma=(0,*template_pixel_sigma),order=0)

    Galaxy_Template['Shifted_Template_Cube'] = Galaxy_Template['Shifted_Template_Cube']*Galaxy_Template['Shifted_Mask']+noise_template*(1.-Galaxy_Template['Shifted_Mask'])
    del noise_template

    del Galaxy_Template['Shifted_Mask']
    #
    final_cube, final_hdr,final_mask = smooth_and_regrid(Galaxy_Template['Shifted_Template_Cube'],
        Galaxy_Template['Shifted_Template_Header'], factor=Galaxy_Template['Shift_Factor'][1], Mask=Galaxy_Template['Final_Mask'],regrid=False)

    del Galaxy_Template['Shifted_Template_Cube']; del Galaxy_Template['Final_Mask']
    del Galaxy_Template['Shifted_Template_Header']
    #For accuracy it would be better to de-rotate before regridding, from an efficiency point this is better
    if final_hdr['BPA'] != 0.:
        final_cube = cf.rotateCube(final_cube,\
        (-1.*final_hdr['BPA']),\
        [final_hdr['CRPIX1'],
        final_hdr['CRPIX2']],order=1)
        final_mask = cf.rotateCube(final_mask,\
        (-1.*final_hdr['BPA']),\
        [final_hdr['CRPIX1'],
        final_hdr['CRPIX2']],order=1)
    #And regrid

        #And remove the shift
    #fits.writeto(f"{galaxy_dir}Test_final.fits", final_cube, final_hdr,
    #             overwrite=True)
    pixel_shift = int(np.sum(Galaxy_Template['Pixel_Buffer']))
    #print(f"This is the pixel shift we remove {pixel_shift} current size {final_cube.shape}")
    final_cube = copy.deepcopy(final_cube[:, \
            pixel_shift:final_hdr['NAXIS2']-pixel_shift,\
            pixel_shift:final_hdr['NAXIS1']-pixel_shift])
    final_mask=  copy.deepcopy(final_mask[:, \
            pixel_shift:final_hdr['NAXIS2']-pixel_shift,\
            pixel_shift:final_hdr['NAXIS1']-pixel_shift])

    final_hdr['NAXIS1'] = final_cube.shape[2]
    final_hdr['NAXIS2'] = final_cube.shape[1]
    final_hdr['CRPIX1'] = final_hdr['CRPIX1'] -pixel_shift
    final_hdr['CRPIX2'] = final_hdr['CRPIX2'] -pixel_shift

    #FINALLY WE REGRID
    pixel_factor = (final_hdr['BMIN'] / 5.) / abs(final_hdr['CDELT2'])
    if pixel_factor >   1.1:
        old_shape = final_cube.shape
        final_cube = cf.regrid_array(final_cube, Out_Shape=((int(final_hdr['NAXIS3']),
            int(final_hdr['NAXIS2'] /  pixel_factor),int(final_hdr['NAXIS1'] / pixel_factor))))
        final_mask = cf.regrid_array(final_mask, Out_Shape=((int(final_hdr['NAXIS3']),
            int(final_hdr['NAXIS2'] /  pixel_factor),int(final_hdr['NAXIS1'] / pixel_factor))))
        final_mask[final_mask > 0.5] = 1.
        final_mask[final_mask <= 0.5] = 0.

        for ext in ['1','2']:
            achieved_regrid = old_shape[int(ext)] / final_cube.shape[int(ext)]
            final_hdr[f'CDELT{ext}'] = final_hdr[f'CDELT{ext}']*achieved_regrid
            final_hdr[f'CRPIX{ext}'] = final_hdr[f'CRPIX{ext}']/achieved_regrid
            final_hdr[f'NAXIS{ext}'] = final_cube.shape[int(ext)]

    #Then shift the header values to the new distance
    final_hdr['CDELT2'] = final_hdr['CDELT2']/Galaxy_Template['Shift_Factor'][0]
    final_hdr['CDELT1'] = final_hdr['CDELT1']/Galaxy_Template['Shift_Factor'][0]
    final_hdr['BMAJ'] = final_hdr['BMAJ']/Galaxy_Template['Shift_Factor'][0]
    final_hdr['BMIN'] = final_hdr['BMIN']/Galaxy_Template['Shift_Factor'][0]
    #And preserve the brightness for the peam change


    Achieved_Noise = np.std(final_cube[0:2,:,:])
    Galaxy_Template['Shifted_TRM_Model']['RMS'] = f"RMS = {str(Achieved_Noise)}"
    Achieved_Mean = cf.get_mean_flux(final_cube,Mask=final_mask)
    Achieved_SNR = Achieved_Mean/Achieved_Noise


    final_hdr['DATAMAX'] = np.max(final_cube)
    final_hdr['DATAMIN'] = np.min(final_cube)
    fits.writeto(f"{galaxy_dir}Convolved_Cube_Gauss.fits", final_cube, final_hdr,
                 overwrite=True)#ANd write to our directory
    final_hdr['DATAMAX'] = np.max(final_mask)
    final_hdr['DATAMIN'] = np.min(final_mask)
    final_mask = np.array(final_mask, dtype=np.int16)
    fits.writeto(f"{galaxy_dir}mask.fits", final_mask, final_hdr,overwrite=True,output_verify='ignore')
    # This completes cube manupulation
    channel_width = final_hdr['CDELT3']
    # the mass leads to the radius at which we need to hit 1 M/pc^2 from Wang (2016) in kpc
    R_HI=[cf.convertskyangle(Galaxy_Template['Final_DHI_arcsec'],distance=\
            Galaxy_Template['Final_Distance'])/2.,0.,0.]
    V_max=float(Galaxy_Template['Shifted_TRM_Model']['VROT'].split('=')[1].split()[-1])
    DynMass=R_HI[0]*10**3*V_max**2/G_agc
    Achieved = {'SNR':Achieved_SNR, 'Mean_Signal': Achieved_Mean, 'Noise': Achieved_Noise,\
            'Channel_Width': channel_width,'Res_Beam':\
            [final_hdr['BMAJ']*3600.,final_hdr['BMIN']*3600.,final_hdr['BPA']],\
            'Pixel_Beam': cf.get_beam_area_in_pixels(final_hdr), 'Mass': DynMass,\
            'Corruption': 'Gaussian' }
    del final_mask ; del final_cube; del final_hdr
    write_overview_file(f"{galaxy_dir}{dirstring}-Info.txt", Galaxy_Template,Achieved,required_noise)
    # We need to make the model input
    with open(f"{galaxy_dir}ModelInput.def", 'w') as tri:
        tri.writelines([Galaxy_Template['Shifted_TRM_Model'][key] + "\n" for key in Galaxy_Template['Shifted_TRM_Model']])

    # And an overview plot
    #print("Start plotting")

    #Profiles are fitted by an exponential with scale length 0.2*RHI i.e. SigHI = C*exp(-R/(0.2*RHI)) so that weay we get the Warp end at Sigma = 0.5
    Warp = [0,np.log(0.5*np.exp(-1./0.2))*-0.2*R_HI[0]]
    cf.plot_input(galaxy_dir,Galaxy_Template['Shifted_TRM_Model']
                ,Title=f'{Galaxy_Template["Name"]} with {Galaxy_Template["Beams"]:.2f} Beams'
                ,Distance= Galaxy_Template['Final_Distance'],RHI=R_HI,WarpR=Warp,
                font_file = font_file )
    # And a file with scrambled initial estimates
    cf.scrambled_initial(galaxy_dir,Galaxy_Template['Shifted_TRM_Model'])
    return_line = f"{Galaxy_Template['Final_Distance']:.2f}|{dirstring}|Convolved_Cube_Gauss\n"
    del Galaxy_Template
    return return_line

create_final_cube.__doc__= f'''
NAME:
   create_final_cube(required_noise, main_directory, Galaxy_Template,font_file = 'empty.ttf')

PURPOSE:
    Add the requested noise and produce the output products

CATEGORY:
   ROC

INPUTS:
    required_noise = the requested noise
    main_directory = 'The main directory to branch from'
    Galaxy_Template = dictionary as created by beam_template
    font_file = location of the font to be used
OPTIONAL INPUTS:

OUTPUTS:
    Cube, ModelInput.def, Initial Estimates and Overview plot.

OPTIONAL OUTPUTS:    exit()

PROCEDURES CALLED:
   Unspecified

NOTE:
'''

def download_templates(galaxy_names,path_to_resources,mp = False, ncpu = 1,\
                        work_dir = './',sofia2 = 'sofia2'):
    if mp:
        #first we check that all templates are processed
        needed_templates = [[x,path_to_resources,work_dir,sofia2] for x in galaxy_names]
        if len(needed_templates) < ncpu:
            ncpu = len(needed_templates)
        with get_context("spawn").Pool(processes=ncpu) as pool:
            results = pool.starmap(check_templates, needed_templates)
        del results
    else:
        for galaxy in galaxy_names:
            result = check_templates(galaxy,path_to_resources,work_dir,sofia2)
        results = []
download_templates.__doc__= f'''
NAME:
   download_templates(galaxy_names,path_to_resources,mp = False, ncpu = 1,\
                           work_dir = './',sofia2 = 'sofia2')
PURPOSE:
    make sure that all templates in galaxy names are present. If not download them.

CATEGORY:
   ROC

INPUTS:
    galaxy_names = names to check
    path_to_resources = path to the pyHIARd installation

OPTIONAL INPUTS:

    mp = False
    Use multiprocessing

    ncpu = 1
    number of cpus to use in multiprocessing

    work_dir = './'
    loacataion where to run sofia

    sofia2 = 'sofia2'
    command to call sofia

OUTPUTS:
    Cube, ModelInput.def, Initial Estimates and Overview plot.

OPTIONAL OUTPUTS:    exit()

PROCEDURES CALLED:
   Unspecified

NOTE:
'''



def extend_cube(cube_in, hdr_in, new_beam=0., Mask=None, rotation_pa=0.):
    if Mask is None:
         Mask=[-1., -1.]
    cube = copy.deepcopy(cube_in)
    hdr = copy.deepcopy(hdr_in)
    Mask_Use = copy.deepcopy(Mask)

    #caculate the buffers
    pixel_buffer = [int(3.*new_beam/abs(hdr['CDELT2']))]
    pixel_buffer.append(int(abs(np.sin(np.radians(rotation_pa))) *\
                    ((hdr['NAXIS1']+pixel_buffer[0])/2.)+3.))

    #Create an extended cube
    extended_cube = np.zeros((hdr['NAXIS3'], hdr['NAXIS2']+2*int(np.sum(pixel_buffer)),\
        hdr['NAXIS2']+2*int(np.sum(pixel_buffer))))
    #and place the cube inside
    extended_cube[:, int(np.sum(pixel_buffer)):hdr['NAXIS2'] + int(np.sum(pixel_buffer)),
        int(np.sum(pixel_buffer)):hdr['NAXIS1'] +int(np.sum(pixel_buffer))] = cube
    #Update the header
    for ax in ['1','2']:
        hdr[f'NAXIS{ax}']= hdr[f'NAXIS{ax}']+2*int(np.sum(pixel_buffer))
        hdr[f'CRPIX{ax}']= hdr[f'CRPIX{ax}']+int(np.sum(pixel_buffer))

    if np.sum(Mask_Use) == -2.:
        return extended_cube,hdr,pixel_buffer
    else:
        extended_mask = np.zeros((hdr['NAXIS3'], Mask_Use.shape[1]+2*int(np.sum(pixel_buffer)),\
            Mask_Use.shape[2]+2*int(np.sum(pixel_buffer))))
        extended_mask[:, int(np.sum(pixel_buffer)):Mask_Use.shape[1]+int(np.sum(pixel_buffer)) ,
            int(np.sum(pixel_buffer)):Mask_Use.shape[2]  +int(np.sum(pixel_buffer))] = Mask_Use
        return extended_cube,hdr,pixel_buffer,extended_mask
extend_cube.__doc__ = f'''
NAME:
     extend_cube(cube_in, hdr_in,new_beam= '0.', Mask = [-1.,-1.])

PURPOSE:
    extend the spatial axes of a cube to create a buffer for rotation or smoothing

CATEGORY:
   roc

INPUTS:
    cube_in = the cube to be extended
    hdr_in = the corrseponding header

OPTIONAL INPUTS:
    new_beam = 0.
    the new beam to accomadate for in degree

    rotation_pa= 0.
    astronomical PA for which the rotation should be accomodated

    Mask = [-1.-1.]
    A cube that should be extended in the same way

OUTPUTS:
    Extended cube and updated header
    pixel_buffer = list of the size of the buffer added [rot_buff,smooth_buffer]

    if both the rotation and beam are set this will be a tuple

OPTIONAL OUTPUTS:

PROCEDURES CALLED:
   Unspecified

NOTE:
'''

def galaxy_template(name,path_to_resources,work_directory,sofia2_call):
    galaxy_module = importlib.import_module(f'pyHIARD.Resources.Cubes.{name}.{name}')
    #Get the template
    Template_All=galaxy_module.get_data(work_directory,sofia2_call)
    Template_Header = Template_All[0].header
    Template_Cube = Template_All[0].data
    Template_All.close()
    print(f"We are processing the template {name}.")
    print(f"We are creating the Original template.")

    #We ensure that the Template is square
    #and the new size of the spatial axes
    new_size = int(np.max(Template_Cube.shape[1:]))
    # create a new area with only zeros
    tmp = np.zeros([Template_Cube.shape[0],new_size,new_size])
    # calculate the shift on each axis
    spatial_extension = []
    for i in [1,2]:
        spatial_extension.append(int((new_size-Template_Cube.shape[i])/2.))
        #Update the header
        Template_Header[f'CRPIX{3-i:d}'] = Template_Header[f'CRPIX{3-i:d}']+spatial_extension[-1]
        Template_Header[f'NAXIS{3-i:d}'] = new_size
    # place the template in our extended cube
    tmp[:,spatial_extension[0]:spatial_extension[0]+Template_Cube.shape[1],\
          spatial_extension[1]:spatial_extension[1]+Template_Cube.shape[2]] = Template_Cube[:,:,:]
    Template_Cube = copy.deepcopy(tmp)
    #close the temporary array
    del tmp
    # If we have a BPA in the observations we want to algin the template such that the BMAJ align with the DEC axis
    if Template_Header['BPA'] != 0.:
        #BPA is anticlockwise while rotate works clockwise so to rotate back means rotate by BPA
        Template_Cube =cf.rotateCube(Template_Cube,\
            (Template_Header['BPA']),\
            [Template_Header['CRPIX1'],
            Template_Header['CRPIX2']],order=1)


    # Obtain the tilted ring model that corresponds to this observation
    Model_Template = get_main_template(name,Template_Header,galaxy_module)
    # Now set a set of values that we required further down the line
    #The systemic velocity of the galaxy
    systemic =  float(Model_Template['VSYS'].split('=')[1].split()[0])
    # We want to ensure that the cube is calibrated on vsys
    Template_Header['CRPIX3'] = (systemic-Template_Header['CRVAL3']) / \
                        Template_Header['CDELT3'] + \
                            Template_Header['CRPIX3']
    #beam area in pixels
    pixels_per_beam = cf.get_beam_area_in_pixels(Template_Header)
    # DHI in arcconds
    DHI_arcsec=cf.convertskyangle(galaxy_module.galaxy_parameters['DHIkpc'],distance=galaxy_module.galaxy_parameters['Distance'],physical=True)

    #print(f"we get originally DHI {DHI_arcsec} arcsec, {galaxy_module.galaxy_parameters['DHIkpc']} kpc, {galaxy_module.galaxy_parameters['Distance']} Mpc")
    #The number of beams in the original observation corresponding to the D_HI diameter

    max_beams_across=(DHI_arcsec)/(Template_Header["BMAJ"]*3600.)
    #!!Maybe we need to place this before the rotate
    Original_Total_Flux_In = np.sum(Template_Cube[Template_Cube > 0.])/pixels_per_beam* Template_Header['CDELT3']#
    # To calculate the mean of all values in the mask is too sensitive to very small variations in the mask
    Original_Mean = cf.get_mean_flux(Template_Cube)
    #print(f'''The mean as determined from the template = {Original_Mean}
#The noise is {galaxy_module.galaxy_parameters["RMS"]}'''    )
    Original_z= np.sqrt((1+systemic/c_kms)/(1-systemic/c_kms))

    Template_Dictionary = {'Name':name,'Galaxy_Template_Cube':Template_Cube,'Galaxy_Template_Header':Template_Header,\
                           'Galaxy_TRM_Model':Model_Template,\
                           'Original_Noise': galaxy_module.galaxy_parameters["RMS"],'Original_Mean_Flux':Original_Mean,\
                           'Original_z':Original_z, 'Original_Vsys':systemic,'Max_Beams_Across':max_beams_across,\
                           'Original_Distance': galaxy_module.galaxy_parameters["Distance"],\
                           'Original_DHI_arcsec': DHI_arcsec, 'Original_Total_Flux_In': Original_Total_Flux_In,\
                           'Disclaimer':galaxy_module.place_disclaimer,\
                           'M_HI':galaxy_module.galaxy_parameters['MHI']}
    return Template_Dictionary
galaxy_template.__doc__=f'''
NAME:
    galaxy_templates(name,beams):

PURPOSE:
    Create all beam templates and the Model Input for a specific galaxy

CATEGORY:
   roc

INPUTS:
    name = basename of template
    beams = the requested beams

OPTIONAL INPUTS:

OUTPUTS:
    boolean to indicate succes

OPTIONAL OUTPUTS:

PROCEDURES CALLED:
   Unspecified

NOTE:
'''


def get_main_template(name,Template_Header,galaxy_module):
    #Open the galaxy's module
    galaxy_resource_path = os.path.dirname(os.path.abspath(galaxy_module.__file__))
    #Check whether we have already created a homogenized template
    homogenized_template_exists = os.path.isfile(f"{galaxy_resource_path}/{name}_uniform.def")
    if homogenized_template_exists:
        #If we have a homogenized template,  just read that stuff
        Template_Model= cf.read_template_file(f"{galaxy_resource_path}/{name}_uniform.def",package_file = False)
    else:
        #If not read the template we have
        radius, rotation, pa,incli,xpos,ypos,systemic,rotation2, pa2,incli2,xpos2,ypos2,systemic2,\
             scaleheight, dispersion, scaleheight2, dispersion2,condisp,sbr,sbr2 = cf.load_text_model(
                name,type =galaxy_module.galaxy_parameters['Original_Model'] ,\
                Variables=['RADI','VROT','PA','INCL','XPOS','YPOS','VSYS','VROT_2',\
                'PA_2','INCL_2','XPOS_2','YPOS_2','VSYS_2','Z0','SDIS','Z0_2',\
                'SDIS_2','CONDISP', 'SBR', 'SBR_2'])
        #Homogenize the input model
        if galaxy_module.galaxy_parameters['Original_Model'] in ['Bar','RC']:
            ndisks= 1
            rotation2= rotation
            pa2=  pa
            incli2= incli
            if galaxy_module.galaxy_parameters['Original_Model'] in ['RC']:
                xpos = [x+Template_Header['CRPIX1'] for x in xpos]
                ypos = [x+Template_Header['CRPIX2'] for x in ypos]
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                input_wcs = WCS(Template_Header).celestial

            xpos,ypos = zip(*[input_wcs.wcs_pix2world(x,y,1) for x,y in zip(xpos,ypos)])
            xpos2=xpos
            ypos2=ypos

        if condisp[0] == -1.:
            condisp = [(Template_Header['CDELT3']*1.2/(2*np.sqrt(2*np.log(2.))))]
        if np.sum(dispersion) == -1.*len(dispersion):
            dispersion[:] = np.sqrt(condisp[0]**2-(Template_Header['CDELT3']*1.2/(2*np.sqrt(2*np.log(2.))))**2)
            dispersion2[:] = np.sqrt(condisp[0]**2-(Template_Header['CDELT3']*1.2/(2*np.sqrt(2*np.log(2.))))**2)
                #continue
        if xpos[0] < 0.:
            xpos = [e+360. for e in xpos]
        if xpos2[0] < 0.:
            xpos2 = [e+360. for e in xpos2]
        RAdeg = xpos[0]
        DECdeg = ypos[0]
        #All values that are in arcsec need to be converted to kpc
        radius_kpc = cf.convertskyangle([x for x in radius if x != -1.],distance=galaxy_module.galaxy_parameters['Distance'])
        if any([x != -1 for x in scaleheight]):
            scaleheight_kpc = cf.convertskyangle([x for x in scaleheight if x != -1.],distance=galaxy_module.galaxy_parameters['Distance'])
        else:
            scaleheight_kpc = [0.]
        if any([x != -1 for x in scaleheight2]):
            scaleheight2_kpc = cf.convertskyangle([x for x in scaleheight2 if x != -1.], distance=galaxy_module.galaxy_parameters['Distance'])
        else:
            scaleheight2_kpc = [0.]
        #Everything that is constant can be written to our the Template def file
        #Read the default template
        Template_Model = cf.read_template_file('Template.def')
        Template_Model['NUR'] = f'NUR = {len(radius_kpc)}'
        Template_Model['RADI'] = f'RADI= {" ".join(str(e) for e in radius_kpc)}'
        Template_Model['INCL'] = 'INCL = '+" ".join(str(e) for e in incli if e != -1.)
        Template_Model['INCL_2'] = 'INCL_2 = ' + " ".join(str(e) for e in incli2 if e != -1.)
        Template_Model['PA'] = 'PA = ' + " ".join(str(e) for e in pa if e != -1.)
        Template_Model['PA_2'] = 'PA_2 = ' + " ".join(str(e) for e in pa2 if e != -1.)
        Template_Model['VROT'] = 'VROT = ' + " ".join(str(e) for e in rotation if e != -1.)
        Template_Model['VROT_2'] = 'VROT_2 = ' + " ".join(str(e) for e in rotation2 if e != -1)
        Template_Model['XPOS'] = 'XPOS = ' + " ".join(str(e) for e in xpos if e != -1. )
        Template_Model['XPOS_2'] = 'XPOS_2 = ' + " ".join(str(e) for e in xpos2 if e != -1.)
        Template_Model['YPOS'] = 'YPOS = ' + " ".join(str(e) for e in ypos if e != -1.)
        Template_Model['YPOS_2'] = 'YPOS_2 = ' + " ".join(str(e) for e in ypos2 if e != -1.)

        if any([x != -1 for x in dispersion]):
            Template_Model['SDIS'] = 'SDIS = ' + " ".join(str(e) for e in dispersion if e != -1.)
        else:
            Template_Model['SDIS'] = 'SDIS =  0.'
        if any([x != -1 for x in dispersion2]):
            Template_Model['SDIS_2'] = 'SDIS_2 = ' + " ".join(str(e) for e in dispersion2 if e != -1.)
        else:
            Template_Model['SDIS_2'] = 'SDIS_2 = 0. '



        if any([x != -1 for x in sbr]):
            Template_Model['SBR'] = 'SBR = ' + " ".join(str(e) for e in sbr if e != -1.)
        else:
            Template_Model['SBR'] = 'SBR =  0.'
        if any([x != -1 for x in sbr2]):
            Template_Model['SBR_2'] = 'SBR_2 = ' + " ".join(str(e) for e in sbr2 if e != -1.)
        else:
            Template_Model['SBR_2'] = 'SBR_2 = 0. '


        Template_Model['Z0'] = f'Z0 = {" ".join([str(e) for e in scaleheight_kpc])}'
        Template_Model['Z0_2'] = f'Z0_2 = {" ".join([str(e) for e in scaleheight2_kpc])}'
        try:
            Template_Header.set("BMAJ", Template_Header["BMMAJ"] / 3600.,before="BMMAJ")
            Template_Header.set("BMIN", Template_Header["BMMIN"] / 3600., before="BMMIN")
        except KeyError:
            pass
            #print("No BMMAJ")
        Template_Model['BMAJ'] = f'BMAJ = {Template_Header["BMAJ"]*3600.}'
        Template_Model['BMIN'] = f'BMIN = {Template_Header["BMIN"]*3600.}'
        Template_Model['BPA'] = f'BPA = {Template_Header["BPA"]}'
        Template_Model['INSET'] = f'INSET = {name}.fits'
        Template_Model['OUTSET'] = f'OUTSET = {name}.fits'
        Template_Model['LOGNAME'] = f'LOGNAME= {name}.log'
        Template_Model['TIRDEF'] = f'TIRDEF= {name}.def'
        Template_Model['CONDISP']= f'CONDISP = {condisp[0]}'
        # As we are now storing this for the main we need to transfer all other values as well
        if any([x != -1 for x in systemic]):
            Template_Model['VSYS'] = f'VSYS = {" ".join([str(e) for e in systemic])}'
        else:
            Template_Model['VSYS'] = f'VSYS = 0'
        if any([x != -1 for x in systemic2]):
            Template_Model['VSYS_2'] = f'VSYS_2 = {" ".join([str(e) for e in systemic2])}'
        else:
            Template_Model['VSYS_2'] = f'VSYS_2 = 0'
        # write this as a homogenized model so we do not have to do this again
        with open(f"{galaxy_resource_path}/{name}_uniform.def",'w') as template:
            for key in Template_Model:
                template.write(f'{Template_Model[key]} \n')
    return Template_Model
get_main_template.__doc__=f'''
NAME:
    get_main_template(name,Template_Header,galaxy_module)

PURPOSE:
    load the main template corresponding to the galaxy

CATEGORY:
   roc

INPUTS:
    name = basename of template
    Template_Header = the header of the template cube. this is required when unifomizing the templates as barollo and rotcur are pixel based
    galaxy_module= the template module

OPTIONAL INPUTS:

OUTPUTS:
    The homogenized template model

OPTIONAL OUTPUTS:

PROCEDURES CALLED:
   Unspecified

NOTE:
'''

def remove_template(galaxy_to_remove,path_to_resources,existing_galaxies):
    existing_galaxies_low = [x.lower() for x in existing_galaxies]
    to_remove = existing_galaxies[existing_galaxies_low.index(galaxy_to_remove.lower())]
    really = cf.get_bool(f"Are you certain you want to remove the template for  {to_remove}? This action can not be undone. (default = yes)",default=True)
    if really:
        print(f"We will remove the template {to_remove}.")
        cf.delete_directory(f"{path_to_resources}{to_remove}")
        if not os.path.isdir(f"{path_to_resources}{to_remove}"):
            print(f"We have succesfully removed the template {to_remove}.")
            return True
        else:
            print(f"Something went wrong. please try again or start an issue on github")
            sys.exit()
    else:
        return False
remove_template.__doc__ = f'''
NAME:
  remove_template(galaxy_to_remove,path_to_resources,existing_galaxies):

PURPOSE:
    remove a galaxy template from the package

CATEGORY:
   roc

INPUTS:
    galaxy_to_remove = Name of the template to remove
    path_to_resources = path to the resources cube directory
    existing_galaxies = list of existing templates

OPTIONAL INPUTS:

OUTPUTS:
    boolean indicating succes or abort

OPTIONAL OUTPUTS:

PROCEDURES CALLED:
   Unspecified

NOTE:
'''

def ROC(cfg,path_to_resources):

    # Let's give an over view of the database that will be created
    print(f"We will create a database with the galaxies {','.join([x for x in cfg.roc.base_galaxies])} basic sets in the directory {cfg.general.main_directory}.\n")
    print("We will vary the following parameters")

    if 'Beams' in cfg.roc.variables_to_vary:
        print("We create the model with {} beams across the major axis.\n".format(", ".join([str(e) for e in cfg.roc.beams])))
    if 'SNR' in cfg.roc.variables_to_vary:
        print("Varying the signal to noise ratio with: {}.\n".format(" ".join([str(e) for e in cfg.roc.snr])))

    Clean_Cube = True
    # If we make new models delete everything in the directory
    if cfg.roc.delete_existing:
        to_delete  =f'''rm -R {' '.join([f"{cfg.general.main_directory}{x}_*Beams_*SNR" for x in cfg.roc.base_galaxies])}'''
        os.system(to_delete)
    Catalogue=f'{cfg.general.main_directory}Output_ROC_Summary.txt'

    number_models= cf.get_created_models(Catalogue,cfg.roc.delete_existing)


    if 'Beams' not in cfg.roc.variables_to_vary:
        cfg.roc.beams= [-1]
    if 'SNR' not in cfg.roc.variables_to_vary:
        cfg.roc.snr= [-1]
    cfg.roc.beams = [float(x) for x in np.sort(cfg.roc.beams)]
    if cfg.roc.beams[0] == -1:
        cfg.roc.beams = np.roll(np.array(cfg.roc.beams),-1).tolist()
    else:
        cfg.roc.beams = list(cfg.roc.beams)
    #if we have a lot of models to create we should split up in chunks

    All_Galaxies = {}
    for name in cfg.roc.base_galaxies:
        #Check whether we need to construct this
        beams_to_produce = []
        combined = []
        for nobeams in cfg.roc.beams:
            noise_to_produce = []
            for reqnoise in cfg.roc.snr:
                dirstring = f"{name}_{nobeams:.1f}Beams_{reqnoise:.1f}SNR"
                beams_to_produce.append(nobeams)
                noise_to_produce.append(reqnoise)
                galaxy_dir =f"{cfg.general.main_directory}{dirstring}/"
                galaxy_dir_exists = os.path.isdir(galaxy_dir)
                if galaxy_dir_exists:
            # Do we have a cube
                    galaxy_cube_exist = os.path.isfile(f"{galaxy_dir}Convolved_Cube_Gauss.fits")

                    if galaxy_cube_exist:
                        beams_to_produce.remove(nobeams)
                        noise_to_produce.remove(reqnoise)
                        print(f"The galaxy {dirstring} galaxy appears fully produced")
            if nobeams not in beams_to_produce:
                print(f"All galaxies with {name} and {nobeams} across the major axis are  thought to be produced already")
            else:
                combined.append([nobeams,noise_to_produce])
        if len(beams_to_produce) > 0:
            All_Galaxies[name] = {'name':name, 'iterations': combined}

    if len(All_Galaxies) == 0:
        print(f"Seems like the ROC has nothing to do.")
        return

    templates = [x for x in All_Galaxies]
    download_templates(templates,path_to_resources, mp = cfg.general.multiprocessing,\
                        ncpu = cfg.general.ncpu,work_dir = cfg.general.main_directory,\
                        sofia2 = cfg.general.sofia2)
    if cfg.general.multiprocessing:
        no_template_processes,no_beam_processes,no_snr_processes = \
            calculate_processes(All_Galaxies,cfg.roc.beams,cfg.roc.snr)
        if  no_snr_processes == 1:
            # if we can not do multiple SNR processes even there is no point in MP
            cfg.general.multiprocessing =False


    if cfg.general.multiprocessing:
        #list with templates to still process
        results = []
        while len(templates) > 0:

            #break templates up to process
            proc_templates = [[x,path_to_resources,cfg.general.main_directory,cfg.general.sofia2] for x in templates[:no_template_processes]]
            processes = cfg.general.ncpu
            if len(proc_templates) < processes:
                processes = len(proc_templates)
            #Then we setup the templates for all beam iterations we wants
            with get_context("spawn").Pool(processes=processes) as pool:
                All_Galaxy_Templates = pool.starmap(galaxy_template, proc_templates)

            del proc_templates
        #All_Galaxy_Templates returns a list of dictionaries that should be complete for constructing all the different beam templates
        #We need to make a list with all the different beams

            different_beams= []
            for galaxy in All_Galaxy_Templates:
                key = galaxy['Name']
                beams_and_noise = [x for x in All_Galaxies[key]['iterations']]
                for x in beams_and_noise:
                    if x[0] > galaxy['Max_Beams_Across']/cfg.roc.minimum_degradation_factor:
                        print(f"You requested {x[0]} beams across. However this galaxy originally only has {galaxy['Max_Beams_Across']} beams across.")
                        print(f"There needs to be a degradation in size for this code to work. Please just use the Original Cube for testing. Or use less than {galaxy['Max_Beams_Across']/cfg.roc.minimum_degradation_factor:.1f} beams across.")
                        print("Continuing to the next beam sizes. If you only want different signal to noise with minimal degradation please set the beams to -1")
                        continue
                    elif x[0] == -1.:
                        #We could not check before whether these exist
                        nobeams = galaxy['Max_Beams_Across']/cfg.roc.minimum_degradation_factor
                        noise_to_produce=[]
                        for reqnoise in cfg.roc.snr:
                            #if reqnoise == -1:
                            #    reqnoise= galaxy['Original_Mean_Flux']/galaxy['Original_Noise']
                            dirstring = f"{key}_{nobeams:.1f}Beams_{reqnoise:.1f}SNR"
                            noise_to_produce.append(reqnoise)
                            galaxy_dir =f"{cfg.general.main_directory}{dirstring}/"
                            galaxy_dir_exists = os.path.isdir(galaxy_dir)
                            if galaxy_dir_exists:
                        # Do we have a cube
                                galaxy_cube_exist = os.path.isfile(f"{galaxy_dir}Convolved_Cube_Gauss.fits")

                                if galaxy_cube_exist:
                                    noise_to_produce.remove(reqnoise)
                                    print(f"The galaxy {dirstring} galaxy appears fully produced")
                        if len(noise_to_produce) == 0:
                            print(f"All galaxies with {key} and {nobeams} across the major axis are  thought to be produced already")
                        else:
                            galaxy['Requested_SNR'] = noise_to_produce
                            different_beams.append((nobeams,galaxy,cfg.roc.max_degradation_factor,cfg.general.main_directory))
                    else:
                        galaxy['Requested_SNR'] = x[1]
                        #if -1. in galaxy['Requested_SNR']:
                        #    reqnoise = galaxy['Original_Mean_Flux']/galaxy['Original_Noise']
                        #    galaxy['Requested_SNR'] = [reqnoise if x == -1. else x for x in galaxy['Requested_SNR']]
                        different_beams.append((x[0],galaxy,cfg.roc.max_degradation_factor,cfg.general.main_directory))
            # if we have more than one template we know that all beams and SNR fit in the pool
            if len(different_beams) == 0:
                print(f"Seems like the ROC has nothing to do on this batch of beam templates.")
                templates = templates[no_template_processes:]
                continue
            del All_Galaxy_Templates

            while len(different_beams) > 0:
                current_different_beams = different_beams[:int(no_beam_processes*no_template_processes)]
                #Now Contruct all the beam templates of the current batch
                processes = cfg.general.ncpu
                if len(current_different_beams) < processes:
                    processes = len(current_different_beams)
                with get_context("spawn").Pool(processes=processes) as pool:
                    All_Beam_Templates = pool.starmap(beam_templates,current_different_beams)
                del current_different_beams
                # now setup a an array with the required noise input
                all_noise = []
                for galaxy in All_Beam_Templates:
                    for value in galaxy['Requested_SNR']:
                        all_noise.append((value,cfg.general.main_directory,galaxy,cfg.general.font_file))

                while len(all_noise) > 0:
                    current_noise= all_noise[:int(no_snr_processes*no_beam_processes*no_template_processes)]
                    processes = cfg.general.ncpu
                    if len(current_noise) < processes:
                        processes = len(current_noise)
                    with get_context("spawn").Pool(processes=processes) as pool:
                        tmp_result = pool.starmap(create_final_cube, current_noise)
                    all_noise = all_noise[int(no_snr_processes*no_beam_processes*no_template_processes):]
                    results = results+tmp_result
                del all_noise
                different_beams = different_beams[int(no_beam_processes*no_template_processes):]

            templates = templates[no_template_processes:]
    else:
        #No MP processing
        results = []
        for key in All_Galaxies:
            galaxy_template_out = galaxy_template(key,path_to_resources,cfg.general.main_directory,cfg.general.sofia2)
            for beam_n_noise in All_Galaxies[key]['iterations']:
                if beam_n_noise[0] > galaxy_template_out['Max_Beams_Across']/cfg.roc.minimum_degradation_factor:
                    print(f"You requested {beam_n_noise[0]} beams across. However this galaxy originally only has {galaxy_template_out['Max_Beams_Across']} beams across.")
                    print(f"There needs to be a degradation in size for this code to work. Please just use the Original Cube for testing. Or use less than {galaxy_template_out['Max_Beams_Across']/cfg.roc.minimum_degradation_factor:.1f} beams across.")
                    print("Continuing to the next beam sizes. If you only want different signal to noise with minimal degradation please set the beams to -1")
                    continue
                elif beam_n_noise[0] == -1.:
                    #We could not check before whether these exist
                    nobeams = galaxy_template_out['Max_Beams_Across']/cfg.roc.minimum_degradation_factor
                    noise_to_produce=[]
                    for reqnoise in cfg.roc.snr:
                        #if reqnoise == -1:
                        #    reqnoise= galaxy['Original_Mean_Flux']/galaxy['Original_Noise']
                        dirstring = f"{key}_{nobeams:.1f}Beams_{reqnoise:.1f}SNR"
                        noise_to_produce.append(reqnoise)
                        galaxy_dir =f"{cfg.general.main_directory}{dirstring}/"
                        galaxy_dir_exists = os.path.isdir(galaxy_dir)
                        if galaxy_dir_exists:
                            # Do we have a cube
                            galaxy_cube_exist = os.path.isfile(f"{galaxy_dir}Convolved_Cube_Gauss.fits")

                            if galaxy_cube_exist:
                                noise_to_produce.remove(reqnoise)
                                print(f"The galaxy {dirstring} galaxy appears fully produced")
                    if len(noise_to_produce) == 0:
                        print(f"All galaxies with {name} and {nobeams} across the major axis are  thought to be produced already")
                        continue
                    else:
                        galaxy_template_out['Requested_SNR'] = noise_to_produce
                        beam_input = [nobeams,galaxy_template_out,cfg.roc.max_degradation_factor,cfg.general.main_directory]
                else:
                    galaxy_template_out['Requested_SNR'] =beam_n_noise[1]
                    beam_input = [beam_n_noise[0],galaxy_template_out,cfg.roc.max_degradation_factor,cfg.general.main_directory]

                beam_template = beam_templates(*beam_input)
                for req_snr in beam_template['Requested_SNR']:
                    result = create_final_cube(float(req_snr),cfg.general.main_directory,beam_template,cfg.general.font_file)
                    results.append(result)


    with open(Catalogue, 'a') as cat:
        for line in results:
            if line != 'EMPTY':
                cat.write(f"{number_models}|{line}")
                number_models += 1
    print(f"The ROC created {number_models} models")





ROC.__doc__ = f'''
NAME:
    ROC

PURPOSE:
    Shift and degrade real galaxies.

CATEGORY:
    agc

INPUTS:
    cfg = OmegaConf Configuration file

OPTIONAL INPUTS:

OUTPUTS:
    A set of real galaxies in a setup ready for FAT fitting

OPTIONAL OUTPUTS:

PROCEDURES CALLED:
   Unspecified

NOTE:
'''

def smooth_and_regrid(Cube_In,hdr_In,factor=1.5,update_header=True, Mask = None,\
                        regrid = True, track_noise = None):
    Cube=copy.deepcopy(Cube_In)
    hdr=copy.deepcopy(hdr_In)
    if Mask is not None:
        Mask_Use = copy.deepcopy(Mask)
    #first calculate the required beams
    new_beam = []
    smooth_beam = []
    sig_pix = []
    for val in ['BMAJ','BMIN']:
        new_beam.append(hdr[val]*factor)
        #Which means we need to smoothe with
        smooth_beam.append(np.sqrt(new_beam[-1] ** 2 -  (hdr[val])** 2))
        # which in pixels is a dispersion of
        if val == 'BMAJ':
            #the major axis is oriented along the declination axis
            pix_size= abs(hdr['CDELT2'])
        else:
            pix_size= abs(hdr['CDELT1'])
        sig_pix.append((smooth_beam[-1] / np.sqrt(8 * np.log(2))) / abs(pix_size))

    if track_noise != None:
        #create a channel of the cube withe the noise of the input
        original_pixel_sigma=(np.array([hdr[key] for key in ['BMAJ','BMIN'] ],dtype=float)/ \
            np.abs(np.array([hdr[key] for key in ['CDELT2','CDELT1'] ],dtype=float)))/\
            (2. * np.sqrt(2. * np.log(2.)))
        achieved_noise,pixel_noise = cf.calculate_pixel_noise(track_noise,original_pixel_sigma,tolerance=0.01)
        #create a cube with this noise
        rng = np.random.default_rng()
        noise_channel = rng.normal(scale=pixel_noise, size=Cube.shape[1:])
        noise_channel = scipy.ndimage.gaussian_filter(noise_channel, sigma=original_pixel_sigma, order=0)
        initial_noise = np.std(noise_channel)
        smoothed_noise = scipy.ndimage.gaussian_filter(noise_channel, sigma=sig_pix, order=0)
        del noise_channel

    smoothed_cube = scipy.ndimage.gaussian_filter(Cube, sigma=(0, *sig_pix), order=0)
    del Cube
    pixel_factor = (new_beam[1] / 5.) / abs(hdr['CDELT2'])
    #Preserve surface brightness
    beams  = [[*new_beam],[hdr['BMAJ'],hdr['BMIN']]]
    new_pix_perbeam = []
    beamar = []
    for beam in beams:
        beamar.append(np.pi*abs((beam[0]*beam[1])**2)/(4.*np.log(2.)))
        new_pix_perbeam.append(cf.get_beam_area_in_pixels(hdr, beam=beam))
    smoothed_cube=smoothed_cube*new_pix_perbeam[0]/new_pix_perbeam[1]
    if track_noise != None:
        smoothed_noise = smoothed_noise *new_pix_perbeam[0]/new_pix_perbeam[1]
    #now regird to 5 pixels across the minor axis
    pixel_factor = (new_beam[1] / 5.) / abs(hdr['CDELT1'])
    if regrid:
        regrid_cube = cf.regrid_array(smoothed_cube, Out_Shape=((int(hdr['NAXIS3']),
            int(hdr['NAXIS2'] /  pixel_factor),int(hdr['NAXIS1'] / pixel_factor))))
        #also the mask Used
        if track_noise != None:
            regrid_noise = cf.regrid_array(smoothed_noise, Out_Shape=((int(hdr['NAXIS2'] /  pixel_factor),int(hdr['NAXIS1'] / pixel_factor))))
            new_noise = np.std(regrid_noise)
            del regrid_noise
        if Mask is not None:
            Mask_Use =  cf.regrid_array(Mask, Out_Shape=((int(hdr['NAXIS3']),
            int(hdr['NAXIS2'] /  pixel_factor),int(hdr['NAXIS1'] / pixel_factor))))
    else:
        regrid_cube = smoothed_cube
        if track_noise != None:
            new_noise = np.std(smoothed_noise)

    # We have to update the header
    if update_header:
        hdr['BMAJ']=new_beam[0]
        hdr['BMIN']=new_beam[1]
        for ext in ['1','2']:
            achieved_regrid = smoothed_cube.shape[int(ext)] / regrid_cube.shape[int(ext)]
            hdr[f'CDELT{ext}'] = hdr[f'CDELT{ext}']*achieved_regrid
            hdr[f'CRPIX{ext}'] = hdr[f'CRPIX{ext}']/achieved_regrid
            hdr[f'NAXIS{ext}'] = regrid_cube.shape[int(ext)]
    if track_noise != None:
        del smoothed_noise
    del smoothed_cube

    return_list  = [regrid_cube,hdr]
    if Mask is not None:
        return_list.append(Mask_Use)
    if track_noise != None:
        return_list.append(track_noise*new_noise/initial_noise)

    return return_list

smooth_and_regrid.__doc__ = f'''
NAME:
    smooth_and_regrid(Cube_In,hdr_In,factor=1.5,update_header=True, Mask = ['EMPTY'])

PURPOSE:
    Shift and degrade real galaxies.

CATEGORY:
    roc

INPUTS:
    Cube_In = Cube to smooth and regrid
    hdr_In = hdr of the input cube
    factor=1.5 factor to smooth the beam by

OPTIONAL INPUTS:
    update_header=True
    booelean to indicate whether the header should be updated

    Mask = ['EMPTY']
    mask that corresponds to cube that should still correspond after regridding.


OUTPUTS:
    A set of real galaxies in a setup ready for FAT fitting

OPTIONAL OUTPUTS:

PROCEDURES CALLED:
   Unspecified

NOTE:
'''

def write_overview_file(filename,Galaxy_Template,Achieved,SNR):
    Template = Galaxy_Template['Shifted_TRM_Model']
    # Then we also want to write some info about the galaxy
    RAdeg = float(Galaxy_Template['Shifted_TRM_Model']['XPOS'].split('=')[1].split()[0])
    DECdeg = float(Galaxy_Template['Shifted_TRM_Model']['YPOS'].split('=')[1].split()[0])
    RAhr, DEChr = cf.convertRADEC(RAdeg, DECdeg)
    RHI = cf.convertskyangle(Galaxy_Template['Final_DHI_arcsec'],float(Template['DISTANCE'].split('=')[1]))/2.
    with open(filename, 'w') as overview:
            overview.write(f'''# This file contains the basic parameters of this galaxy. For the radial dependencies look at Overview.png or ModelInput.def.
#{'Variable':<14s} {'Requested':<15s} {'Achieved':15s} {'Units':<15s}
{'Inclination':<15s} {'':<15s} {float(Template['INCL'].split('=')[1].split()[0]):<15.2f} {'degree':<15s}
{'PA':<15s} {'':<15s} {float(Template['PA'].split('=')[1].split()[0]):<15.2f} {'degree':<15s}
{'Sys. Velocity':<15s} {'':<15s} {float(Template['VSYS'].split('=')[1].split()[0]):<15.2f} {'km/s':<15s}
{'RA':<15s} {'':<15s} {RAhr.strip():<15s} {'':<15s}
{'Declination':<15s} {'':<15s} {DEChr.strip():<15s} {'':<15s}
{'Dispersion':<15s} {'':<15s} {f"{float(Template['SDIS'].split('=')[1].split()[0]):.2f}-{float(Template['SDIS'].split('=')[1].split()[-1]):.2f}":<15s} {'km/s':<15s}
{'Scale height':<15s} {'':<15s} {f"{float(Template['Z0'].split('=')[1].split()[0]):.2f}-{float(Template['Z0'].split('=')[1].split()[-1]):.2f}":<15s} {'arcsec':<15s}
{'HI Radius':<15s} {'':<15s} {RHI:<15.3f} {'kpc':<15s}
{'Maj Axis':<15s} {Galaxy_Template['Beams']:<15.3f} {'':15s} {'Beams'}
{'Total Mass':<15s} {'':<15s} {Achieved['Mass']:<15.1e} {'M_solar':<15s}
{'HI Mass':<15s} {'':<15s} {Galaxy_Template['M_HI']:<15.1e} {'M_solar':<15s}
{'Chan. Width':<15s} {'':<15s} {Achieved['Channel_Width']:<15.3f} {'km/s':<15.2s}
{'SNR':<15s} {SNR:<15.3f} {Achieved['SNR']:<15.3f}
{'Mean Signal':<15s} {'':<15s} {Achieved['Mean_Signal']*1000.:<15.3f} {'mJy/beam':<15s}
{'Noise':<15s} {'':<15s} {Achieved['Noise']*1000.:<15.3f} {'mJy/beam':<15s}
{'Distance':<15s} {'':<15s} {float(Template['DISTANCE'].split('=')[1]):<15.3f} {'Mpc':<15s}
{'Maj. FWHM':<15s} {'':<15s} {Achieved['Res_Beam'][0]:<15.2f} {'arcsec':<15s}
{'Min. FWHM':<15s} {'':<15s} {Achieved['Res_Beam'][1]:<15.2f} {'arcsec':<15s}
{'Beam BPA':<15s} {'':<15s} {Achieved['Res_Beam'][2]:<15.2f} {'degree':<15s}
{'Pixel per Beam':<15s} {'':<15s} {Achieved['Pixel_Beam']:<15.2f} {'pixel':<15s}
{'Corruption':<15s} {'':<15s} {Achieved['Corruption']:<15s}''')

write_overview_file.__doc__ = f'''
NAME:
   write_overview_file

PURPOSE:
    write the overview file with requested and achieved values

CATEGORY:
   ROC

INPUTS:
   Current_Galaxy = the Current Gakaxy that is processed (The requested values)
   Template = The tirific template that is used to create the galaxy
   Achieved= Copy of Current_Galaxy that contains the values measured from the cube

OPTIONAL INPUTS:

OUTPUTS:
    file with a table with the requested and achieved values
OPTIONAL OUTPUTS:

PROCEDURES CALLED:
   Unspecified

NOTE: For now the Achieved.Flare and Achieved.Warp are merely copies of the input simply assumed to be implemented correctly
'''





from pyHIARD.common_functions import download_cube,create_masks,select_emission
from astropy.io import fits
import os

galaxy_parameters = {'Galaxy': 'NGC_5204', 'DHIkpc': 9.6   ,'Distance': 4.1, 'Original_Model': 'Tir', 'RMS': 0.00038  , 'MHI': 0.54e9  }


class PackageError(Exception):
    pass
def get_data(work_dir,sofia_call='sofia2'):
    '''Download the data for this galaxy and prepare the cube for usage'''
    succes= False
    outdir = os.path.dirname(os.path.abspath(__file__))
    try:
        Cube = fits.open(f"{outdir}/NGC_5204.fits",uint = False, do_not_scale_image_data=True,ignore_blank = True)
    except FileNotFoundError:
        url = ''
        name = 'NGC_5204'
        sizes=[[12,58],[15,180],[20,180]]
        try:
            Cube = fits.open(f"{outdir}/{name}_Original.fits",uint = False, do_not_scale_image_data=True,ignore_blank = True)
        except:
            raise PackageError(f'''This file should been downloaded with the install.
Please list an issue on the Github.''')
        Clean_Cube,hdr = select_emission(Cube[0].data,Cube[0].header,name,work_dir,sofia_call=sofia_call)
        fits.writeto(f"{outdir}/{name}.fits",Clean_Cube,hdr,overwrite = False)
        Cube[0].data=Clean_Cube
        Cube[0].header=hdr
        del Clean_Cube
        del hdr
    #place_disclaimer(dir_to_place)
    return Cube
get_data.__doc__=f'''
NAME:
   get_data

PURPOSE:
   Download the data for this galaxy and prepare the cube for usage

CATEGORY:
   agc

INPUTS:

OPTIONAL INPUTS:


OUTPUTS:
   Cube = Cube in astropy format

OPTIONAL OUTPUTS:

PROCEDURES CALLED:
   Unspecified

NOTE:
'''

def place_disclaimer(dir_to_place):
    disclaimer = '''------ NGC 5204 -----
    Is observed and modelled with TiRiFiC by P. G.I.G. Jozsa.
    If you use this cube for scientific analysis please acknowledge it through a reference to Jozsa et al. 2007
    http://adsabs.harvard.edu/abs/2007A%26A...468..903J

    Original name n5204_lo.fits

    --------------------------------------------------------------
'''
    with open(f'{dir_to_place}/ACKNOWLEDGE.txt', 'w') as file:
        file.writelines(disclaimer)
place_disclaimer.__doc__=f'''
NAME:
   place_disclaimer

PURPOSE:
   Place a disclaimer about the source, models and acknowledgements in the directory.

CATEGORY:
   agc

INPUTS:
    dir_to_place = The directory where the galaxy is cretated.

OPTIONAL INPUTS:


OUTPUTS:

OPTIONAL OUTPUTS:

PROCEDURES CALLED:
   Unspecified

NOTE:
'''



#!/usr/local/bin/ python3
# This is the main program to easily create the real and false database

from pyHIARD.AGC.AGC import AGC
from pyHIARD.ROC.ROC import ROC
from pyHIARD.conf.config import setup_conf


def main_with_trace():
    from viztracer import VizTracer
    with VizTracer(output_file="pyHIARD_Viztracer.json",min_duration=1000) as tracer:
        main()

def main():
    cfg,path_to_resources = setup_conf()
    if cfg.agc.enable:
        AGC(cfg)
    if cfg.roc.enable:
        ROC(cfg,path_to_resources)
    Catalogue = f'{cfg.general.main_directory}/Output_Summary.txt'
    # If we are making new models we want to ensure this is a new file
    with open(Catalogue, 'w') as cat:
        cat.write('ID|Distance|Directoryname|Cubename\n')
        if cfg.agc.enable:
            with open(
                f'{cfg.general.main_directory}/Output_AGC_Summary.txt', 'r') as cat_AGC:
                AGCline = cat_AGC.readlines()
                for line in AGCline:
                    tmp = line.split('|')
                    try:
                        lastnum = int(tmp[0])
                    except:
                        continue
                    cat.write(line)
                lastnum += 1
        else:
            lastnum = 0.
        if cfg.roc.enable:
            with open(
                f'{cfg.general.main_directory}/Output_ROC_Summary.txt', 'r') as cat_ROC:
                ROCline = cat_ROC.readlines()
                for line in ROCline:
                    tmp = line.split('|')
                    try:
                        l = int(tmp[0])
                    except:
                        continue
                    cat.write(
                        str(int(lastnum+int(tmp[0])))+'|'+tmp[1]+'|'+tmp[2]+'|'+tmp[3])
                lastroc = int(tmp[0])+1
        else:
            lastroc = 0
    print(f"In this pyHIARD run we have created {lastnum} AGC models and {lastroc} ROC cubes. In total this makes this database {int(lastnum+lastroc)} models big")





from pyHIARD.common_functions import download_cube,create_masks,select_emission
from astropy.io import fits
import os

galaxy_parameters = {'Galaxy': 'NGC_5023', 'DHIkpc': 18.6  ,'Distance': 6.6, 'Original_Model': 'Tir', 'RMS': 0.0002 , 'MHI': 6.1e8  }


def get_data(work_dir,sofia_call='sofia2'):
    '''Download the data for this galaxy and prepare the cube for usage'''
    succes= False
    outdir = os.path.dirname(os.path.abspath(__file__))
    try:
        Cube = fits.open(f"{outdir}/NGC_5023.fits",uint = False, do_not_scale_image_data=True,ignore_blank = True)
    except FileNotFoundError:
        url = 'https://zenodo.org/record/3715549/files/NGC5023-HR-cube.fits'
        name = 'NGC_5023'
        sizes=[[7,62],[420,600],[430,570]]
        try:
            Cube = fits.open(f"{outdir}/{name}_Original.fits",uint = False, do_not_scale_image_data=True,ignore_blank = True)
        except:
            Cube = download_cube(f'{name}_Original',url,sizes,outdir)
        Clean_Cube,hdr = select_emission(Cube[0].data,Cube[0].header,name,work_dir,sofia_call=sofia_call)
        fits.writeto(f"{outdir}/{name}.fits",Clean_Cube,hdr,overwrite = False)
        Cube[0].data=Clean_Cube
        Cube[0].header=hdr
        if url != '':
            os.system(f"rm -f {outdir}/{name}_Original.fits")
        del Clean_Cube
        del hdr
    #place_disclaimer(dir_to_place)
    return Cube
get_data.__doc__=f'''
NAME:
   get_data

PURPOSE:
   Download the data for this galaxy and prepare the cube for usage

CATEGORY:
   agc

INPUTS:

OPTIONAL INPUTS:


OUTPUTS:
   Cube = Cube in astropy format

OPTIONAL OUTPUTS:

PROCEDURES CALLED:
   Unspecified

NOTE:
'''

def place_disclaimer(dir_to_place):
    disclaimer = '''----- NGC 5023 -----

    is part of the HALOGAS Survey (Heald et al. 2011)
    If you make use of these data products in any publication or presentation, we kindly ask you to cite the following paper(s):

    Primary HALOGAS reference: Heald et al. (2011)
    and to include the following acknowledgement:

    This research made use of data from WSRT HALOGAS-DR1. The Westerbork Synthesis Radio Telescope is operated by ASTRON (Netherlands Institute for Radio Astronomy) with support from the Netherlands Foundation for Scientific Research NWO.

    Please download the original HR cube from
    https://www.astron.nl/halogas/data.php

    Original Cube Name  NGC5023-HR-cube.fits

    This model is created by P.Kamphuis and published in
    HALOGAS observations of NGC 5023 and UGC 2082: Modeling of non-cylindrically symmetric gas distributions in edge-on galaxies" Kamphuis, P.; Rand, R. J.; Józsa, G. I. G.; Zschaechner, L. K.; Heald, G. H.; Patterson, M. T.; Gentile, G.; Walterbos, R. A. M.; Serra, P.; & de Blok, W. J. G. 2013, MNRAS, 434, 2069

'''
    with open(f'{dir_to_place}/ACKNOWLEDGE_HALOGAS.txt', 'w') as file:
        file.writelines(disclaimer)
place_disclaimer.__doc__=f'''
NAME:
   place_disclaimer

PURPOSE:
   Place a disclaimer about the source, models and acknowledgements in the directory.

CATEGORY:
   agc

INPUTS:
    dir_to_place = The directory where the galaxy is cretated.

OPTIONAL INPUTS:


OUTPUTS:

OPTIONAL OUTPUTS:

PROCEDURES CALLED:
   Unspecified

NOTE:
'''



#!/usr/local/bin/ python3
# This module contains a set of functions and classes that are used in several different Python scripts in the Database.

from astropy.utils.data import download_file, clear_download_cache
from astropy.io import fits
from collections import OrderedDict  # used in Proper_Dictionary
from pyHIARD import Templates as templates
from pyHIARD.AGC.base_galaxies import Base_Galaxy
from pyHIARD.Resources import Cubes as cubes
from scipy.ndimage import gaussian_filter, rotate, zoom,map_coordinates

import copy  # Used in columndensities
import numpy as np  # Used in convertskyangle and columndensity and
import os
import pyHIARD
import re
import resource
import signal
import subprocess
import sys
import time
import traceback
import warnings

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import matplotlib
    matplotlib.use('pdf')
    import matplotlib.pyplot as plt
    import matplotlib.font_manager as mpl_fm

try:
    import importlib.resources as import_res
except ImportError:
    import importlib_resources as import_res
 


class SofiaFaintError(Exception):
    pass
class InputError(Exception):
    pass
class CasaInstallError(Exception):
    pass

class SofiaRunError(Exception):
    pass
# A class of ordered dictionary where keys can be inserted in at specified locations or at the end.


class Proper_Dictionary(OrderedDict):
    def insert(self, existing_key, new_key, key_value):
        done = False
        if new_key in self:
            self[new_key] = key_value
            done = True
        else:
            new_orderded_dict = self.__class__()
            for key, value in self.items():
                new_orderded_dict[key] = value
                if key == existing_key:
                    new_orderded_dict[new_key] = key_value
                    done = True
            if not done:
                new_orderded_dict[new_key] = key_value
                done = True
                print(
                    "----!!!!!!!! YOUR new key was appended at the end as you provided a non-existing key to add it after!!!!!!---------")
            self.clear()
            self.update(new_orderded_dict)

        if not done:
            print("----!!!!!!!!We were unable to add your key!!!!!!---------")
#Function to convert column densities
def ask_main_directory(current_directory,input_directory = 'NonE'):
    while not os.path.isdir(input_directory):
        if not input_directory == 'NonE':
                print(f'''The directory {input_directory} does not exist.
Please provide the correct directory''')
        input_directory =  input(\
            f'''Please provide the directory where to create the database.
(default = {current_directory}):''')
        if input_directory == '':
            input_directory = current_directory
    return input_directory

def check_main_directory(cfg):
    current_directory = os.getcwd()
    if cfg.general.main_directory != current_directory:
        correct_directory = get_bool(f'''!!! Your creation directory is not directory from which you started pyHIARD.
Are you sure you want to create the database in:
{cfg.general.main_directory}?   !!!!!!
Please type yes or no (default = yes) ''',default = True)
        if not correct_directory:
            cfg.general.main_directory = ask_main_directory(current_directory)
   
    #Check the main directory exists
    while not os.path.isdir(cfg.general.main_directory):
        cfg.general.main_directory = ask_main_directory(current_directory,\
                                        input_directory=cfg.general.main_directory )
        #if we want the full database default we check that the user wants this
    if cfg.general.main_directory[-1] != '/':
        cfg.general.main_directory = f"{cfg.general.main_directory}/"
    return cfg

def create_directory(directory, base_directory, debug=False):
    split_directory = [x for x in directory.split('/') if x]
    split_directory_clean = [x for x in directory.split('/') if x]
    split_base = [x for x in base_directory.split('/') if x]
    #First remove the base from the directory but only if the first directories are the same
    if split_directory[0] == split_base[0]:
        for dirs, dirs2 in zip(split_base, split_directory):
            if dirs == dirs2:
                split_directory_clean.remove(dirs2)
            else:
                if dirs != split_base[-1]:
                    raise InputError(
                        f"You are not arranging the directory input properly ({directory},{base_directory}).")
    for new_dir in split_directory_clean:
        if not os.path.isdir(f"{base_directory}/{new_dir}"):
            os.mkdir(f"{base_directory}/{new_dir}")
        base_directory = f"{base_directory}/{new_dir}"


create_directory.__doc__ = f'''
 NAME:
    create_directory

 PURPOSE:
    create a directory recursively if it does not exists and strip leading directories when the same fro the base directory and directory to create

 CATEGORY:
    support_functions

 INPUTS:
    directory = string with directory to be created
    base_directory = string with directory that exists and from where to start the check from

 OPTIONAL INPUTS:
    debug = False

 OUTPUTS:

 OPTIONAL OUTPUTS:
    The requested directory is created but only if it does not yet exist

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''


def check_input(cfg):
    cfg = check_main_directory(cfg)

    # if we only have a single cpu turn multiprocessing of
    if cfg.general.ncpu == 1:
        cfg.general.multiprocessing = False

    if cfg.agc.enable:
        cfg.general.tirific = find_program(cfg.general.tirific, 'TiRiFiC')

        #    cfg.general.casa = find_program(cfg.general.casa, 'CASA')
    if cfg.roc.enable:
        cfg.general.sofia2 = find_program(cfg.general.sofia2, 'SoFiA2')

    if cfg.agc.enable:

        #sets = 5  # This is the amount of base galaxies we want, i.e. the number of rotation curves
        question_base = False
        for gals in cfg.agc.base_galaxies:
            if not 1 <= gals <= 7:
                question_base = True
        if question_base:
            print(f'''Please select the base types you would like to use''')
            for i in range(1, 6):
                print(
                    f'''{i}) Galaxy {i} has the following Base parameters to vary on.''')
                print_base_galaxy(Base_Galaxy(i))

            vals = input(
                f"Or you can construct you own by selecting 7, default = 7: ")
            if vals == '':
                cfg.agc.base_galaxies = [7]
            else:
                cfg.agc.base_galaxies = [int(x) for x in re.split(
                    "\s+|\s*,\s*|\s+$", vals.strip()) if 1 <= int(x) <= 6]
                if len(cfg.agc.base_galaxies) == 0:
                    cfg.agc.base_galaxies = [7]
        while cfg.agc.corruption_method.lower() not in ['casa_sim', 'gaussian','no_corrupt', 'casa_5','tres']:
            cfg.agc.corruption_method = input(
                'Your method of corruption is not acceptable please choose from Casa_Sim, Gaussian, No_Corrupt, Tres, Casa_5 (Default = Gaussian):')
            if cfg.agc.corruption_method == '':
                cfg.agc.corruption_method = 'Gaussian'

        if cfg.agc.corruption_method.lower() == 'casa_sim' or cfg.agc.corruption_method.lower() == "c_s":
            cfg.agc.corruption_method = 'Casa_Sim'
            print(
                "You are using the casa corruption method please make sure python can access casa.")
        elif cfg.agc.corruption_method .lower() == 'gaussian' or cfg.agc.corruption_method == "" or cfg.agc.corruption_method .lower() == 'g':
            cfg.agc.corruption_method = 'Gaussian'
        elif cfg.agc.corruption_method .lower() == 'casa_5' or cfg.agc.corruption_method .lower() == "c_5":
            cfg.agc.corruption_method = 'Casa_5'
            print(
                "You are using the casa corruption method please make sure python can access casa.")
        channel_options = ['independent', 'sinusoidal', 'hanning']
        while cfg.agc.channel_dependency.lower() not in channel_options:
            cfg.agc.channel_dependency = input(f''' {cfg.agc.channel_dependency} is not a valid channel dependency
Please choose from {','.join([x for x in channel_options])}:''')

        if cfg.agc.corruption_method.lower() in ['casa_sim', 'casa_5', 'tres']:
            try:
                from casaconfig import AutoUpdatesNotAllowed
                import casatasks
            except ModuleNotFoundError:
                if pyHIARD.__casa_max__ != 'OK':
                    raise CasaInstallError(f'''Your modular casa is not installed.
Most likely because it is not available for python version {sys.version}
As such you can not run the corrupt casa method''')
                else:
                    raise CasaInstallError(f'''Your modular casa is not installed.
We do not know the reason for this but it means you cannot run the casa corrupt method.
''')
            except AutoUpdatesNotAllowed:
                raise CasaInstallError(f'''Since Casa 6.6.6 Casa wants to update its data in ~/.casa/data
that directory does not exist. Please create it or see the Casa website for other options.
Alternatively you can run pyHIARD agc.corruption_method = No_Corrupt or Gauss
''')

              

        #question_variations = False
        changes_poss = ['Inclination', 'PA', 'Beams', 'Radial_Motions', 'Flare', 'Arms',
            'Dispersion', 'Bar', 'Channelwidth', 'SNR', 'Warp', 'Mass', 'Beam_Size', 'Base']
        changes_poss_lower = [x.lower() for x in changes_poss]
        for i, variables in enumerate(cfg.agc.variables_to_vary):
            while variables.lower() not in changes_poss_lower:
                variables = input(f'''{variables} is not a legitimate variable to variations.
Please choose from {','.join([x for x in changes_poss])}
replace with:''')

            cfg.agc.variables_to_vary[i] = changes_poss[changes_poss_lower.index(
                variables.lower())]
        #We always make the base
        #if 'Base' not in cfg.agc.variables_to_vary:
        #    cfg.agc.variables_to_vary.append('Base')

        if 'inclination' in [x.lower() for x in cfg.agc.variables_to_vary]:
            for i, incs in enumerate(cfg.agc.inclination):
                while not 0. <= incs <= 90.:
                    incs = float(
                        input(f'please choose a value between 0.- 90.:'))
                cfg.agc.inclination[i] = incs
        if 'pa' in [x.lower() for x in cfg.agc.variables_to_vary]:
            for i, incs in enumerate(cfg.agc.pa):
                while not 0. <= incs <= 360.:
                    incs = float(
                        input(f'please choose a value between 0.- 360.:'))
                cfg.agc.pa[i] = incs
        #the double lists to be proper, OmegaConf will throw a error when list is not list
        list_to_check = ['warp','dispersion','beam_size']
        elements = [2,2,3]
        for var_to_check in list_to_check:
            if var_to_check in [x.lower() for x in cfg.agc.variables_to_vary]:
                current =getattr(cfg.agc,var_to_check)
                try:
                    if len(current[0]) == elements[list_to_check.index(var_to_check)]:
                        pass
                except TypeError:
                    if len(current) == elements[list_to_check.index(var_to_check)]:
                        #We assume that the user wanted a single variation
                        setattr(cfg.agc,var_to_check,[[x for x in current]])
                    else:
                        raise  InputError(f'''You wanted variations in {var_to_check} which should have {elements[list_to_check.index(var_to_check)]} per variation.
                        Your list has {len(current)} elements and we do not know what to do with that. Note that this is a double list, i.e. [[]]''')


    if cfg.roc.enable:

        path_to_resources = os.path.dirname(os.path.abspath(cubes.__file__))
        allowed_galaxies = [name for name in os.listdir(
            path_to_resources) if os.path.isdir(os.path.join(path_to_resources, name))]
        allowed_galaxies.remove('__pycache__')
        allowed_galaxies_low = [x.lower() for x in allowed_galaxies]
        for i, galaxy in enumerate(cfg.roc.base_galaxies):
            while galaxy.lower() not in allowed_galaxies_low:
                galaxy = input(f'''{galaxy} is not a legitimate galaxy to vary.
Please choose from {','.join([x for x in allowed_galaxies])}
replace with:''')
            cfg.roc.base_galaxies[i] = allowed_galaxies[allowed_galaxies_low.index(
                galaxy.lower())]
        changes_poss = ['Beams', 'SNR']
        changes_poss_lower = [x.lower() for x in changes_poss]
        for i, variables in enumerate(cfg.roc.variables_to_vary):
            while variables.lower() not in changes_poss_lower:
                variables = input(f'''{variables} is not a legitimate variable to variations.
Please choose from {','.join([x for x in changes_poss])}
replace with:''')

            cfg.roc.variables_to_vary[i] = changes_poss[changes_poss_lower.index(
                variables.lower())]
        if cfg.roc.minimum_degradation_factor < 1.1:
            print(f'''Your minimum degration factor is less than 1.1 ({cfg.roc.minimum_degradation_factor}).
That wil create a mess. We will abort this run.''')
            raise InputError(f'''Your minimum degradation factor is too small''')


        if cfg.roc.minimum_degradation_factor > cfg.roc.max_degradation_factor:
            print(f'''Your max degradation factor is less than the mimimum ({cfg.roc.max_degradation_factor} vs { cfg.roc.minimum_degradation_factor}).
We have equalized them for you. Your welcome.''')
            cfg.roc.max_degradation_factor = cfg.roc.minimum_degradation_factor
    if cfg.roc.enable and cfg.roc.delete_existing:
        to_delete = f'''rm -R {' '.join([f"{cfg.general.main_directory}{x}_*Beams_*SNR" for x in cfg.roc.base_galaxies])}'''
        print("All previous models of the requested base galaxies will be removed prior to the build. \n")

        print(f"The command {to_delete} will be run.")
        cfg.roc.delete_existing = get_bool(
            "Are you sure you want to do this? (Yes/No, default=No): ", default=False)

    return cfg


check_input.__doc__ = f'''
 NAME:
    check_input

 PURPOSE:
    Check the configuration input and that all required input is present.
    In case of missing input request the input from the user.

 CATEGORY:
    common_functions

 INPUTS:
    cfg = omega conf configuration

 OPTIONAL INPUTS:

 OUTPUTS:
    a modified cfg

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''


def unused_columndensity(levels, systemic=100., beam=None, channel_width=1., \
                column=False, arcsquare=False, solar_mass=False):
    if beam is None:
        beam=[1., 1.]
    #set solar_mass to indicate the output should be M_solar/pc**2 or if column = True the input is
    f0 = 1.420405751786E9  # Hz rest freq
    c = 299792.458  # light speed in km / s
    pc = 3.086e+18  # parsec in cm
    solarmass = 1.98855e30  # Solar mass in kg
    mHI = 1.6737236e-27  # neutral hydrogen mass in kg

    if systemic > 10000:
        systemic = systemic/1000.
    f = f0 * (1 - (systemic / c))  # Systemic frequency
    if arcsquare:
        HIconv = 605.7383 * 1.823E18 * (2. * np.pi / (np.log(256.)))
        if column:
            # If the input is in solarmass we want to convert back to column densities
            if solar_mass:
                levels = levels*solarmass/(mHI*pc**2)
            #levels=levels/(HIconv*channel_width)
            levels = levels/(HIconv*channel_width)
        else:

            levels = HIconv*levels*channel_width
            if solar_mass:
                levels = levels/solarmass*(mHI*pc**2)
    else:
        if beam.size < 2:
            beam = [beam, beam]
        b = beam[0]*beam[1]
        if column:
            if solar_mass:
                levels = levels*solarmass/(mHI*pc**2)
            TK = levels/(1.823e18*channel_width)
            levels = TK/(((605.7383)/(b))*(f0/f)**2)
        else:
            TK = ((605.7383)/(b))*(f0/f)**2*levels
            levels = TK*(1.823e18*channel_width)
    if ~column and solar_mass:
        levels = levels*mHI*pc**2/solarmass
    return levels
        # a Function to convert the RA and DEC into hour angle (invert = False) and vice versa (default)


def convertRADEC(RA, DEC, invert=False, colon=False):

    if not invert:
        try:
            _ = (e for e in RA)
        except TypeError:
            RA = [RA]
            DEC = [DEC]
        for i in range(len(RA)):
            xpos = RA
            ypos = DEC
            xposh = int(np.floor((xpos[i]/360.)*24.))
            xposm = int(np.floor((((xpos[i]/360.)*24.)-xposh)*60.))
            xposs = (((((xpos[i]/360.)*24.)-xposh)*60.)-xposm)*60
            yposh = int(np.floor(np.absolute(ypos[i]*1.)))
            yposm = int(np.floor((((np.absolute(ypos[i]*1.))-yposh)*60.)))
            yposs = (((((np.absolute(ypos[i]*1.))-yposh)*60.)-yposm)*60)
            sign = ypos[i]/np.absolute(ypos[i])
            if colon:
                RA[i] = "{}:{}:{:2.2f}".format(xposh, xposm, xposs)
                DEC[i] = "{}:{}:{:2.2f}".format(yposh, yposm, yposs)
            else:
                RA[i] = "{}h{}m{:2.2f}".format(xposh, xposm, xposs)
                DEC[i] = "{}d{}m{:2.2f}".format(yposh, yposm, yposs)
            if sign < 0.: DEC[i] = '-'+DEC[i]
        if len(RA) == 1:
            RA = str(RA[0])
            DEC = str(DEC[0])
    else:
        if isinstance(RA, str):
            RA = [RA]
            DEC = [DEC]

        xpos = RA
        ypos = DEC

        for i in range(len(RA)):
            # first we split the numbers out
            tmp = re.split(r"[a-z,:]+", xpos[i])
            RA[i] = (float(tmp[0])+((float(tmp[1])+(float(tmp[2])/60.))/60.))*15.
            tmp = re.split(r"[a-z,:'\"]+", ypos[i])
            DEC[i] = float(np.absolute(float(tmp[0]))+((float(tmp[1])
                           + (float(tmp[2])/60.))/60.))*float(tmp[0])/np.absolute(float(tmp[0]))

        if len(RA) == 1:
            RA = float(RA[0])
            DEC = float(DEC[0])
    return RA, DEC


# function for converting kpc to arcsec and vice versa

def convertskyangle(angle, distance=1., unit='arcsec', distance_unit='Mpc', \
                        physical=False):
    try:
        _ = (e for e in angle)
    except TypeError:
        angle = [angle]

        # if physical is true default unit is kpc
    angle = np.array(angle)
    if physical and unit == 'arcsec':
        unit = 'kpc'
    if distance_unit.lower() == 'mpc':
        distance = distance * 10 ** 3
    elif distance_unit.lower() == 'kpc':
        distance = distance
    elif distance_unit.lower() == 'pc':
        distance = distance / (10 ** 3)
    else:
        print('CONVERTSKYANGLE: ' + distance_unit
              + ' is an unknown unit to convertskyangle.\n')
        print('CONVERTSKYANGLE: please use Mpc, kpc or pc.\n')
        sys.exit()
    if not physical:
        if unit.lower() == 'arcsec':
            radians = (angle / 3600.) * ((2. * np.pi) / 360.)
        elif unit.lower() == 'arcmin':
            radians = (angle / 60.) * ((2. * np.pi) / 360.)
        elif unit.lower() == 'degree':
            radians = angle * ((2. * np.pi) / 360.)
        else:
            print('CONVERTSKYANGLE: ' + unit
                  + ' is an unknown unit to convertskyangle.\n')
            print('CONVERTSKYANGLE: please use arcsec, arcmin or degree.\n')
            sys.exit()

        kpc = 2. * (distance * np.tan(radians / 2.))
    else:
        if unit.lower() == 'kpc':
            kpc = angle
        elif unit.lower() == 'mpc':
            kpc = angle * (10 ** 3)
        elif unit.lower() == 'pc':
            kpc = angle / (10 ** 3)
        else:
            print('CONVERTSKYANGLE: ' + unit
                  + ' is an unknown unit to convertskyangle.\n')
            print('CONVERTSKYANGLE: please use kpc, Mpc or pc.\n')
            sys.exit()
        radians = 2. * np.arctan(kpc / (2. * distance))
        kpc = (radians * (360. / (2. * np.pi))) * 3600.
    if len(kpc) == 1:
        kpc = float(kpc[0])
    return kpc

def select_emission(data,hdr,name,work_dir,sofia_call='sofia_call'):
    #get a mask
    Mask_Data = create_masks(data,hdr, work_dir, name, sofia_call=sofia_call)
    try:
        Vel_Units = hdr['CUNIT3'].lower()
    except KeyError:
        if hdr['CDELT3'] > 150.:
            hdr.set("CUNIT3", 'M/S', before="CTYPE3")
        else:
            hdr.set("CUNIT3", 'KM/S', before="CTYPE3")
    if hdr['CUNIT3'].lower() == 'm/s' or hdr['CDELT3'] > 150.:
        hdr['CDELT3'] = hdr['CDELT3'] / 1000.
        hdr['CUNIT3'] = 'km/s'
        hdr['CRVAL3'] = hdr['CRVAL3'] / 1000.
    # ensure we have a BPA
    if 'BPA' not in hdr:
        hdr['BPA'] = 0.
    #Apply the mask to Template
    data[Mask_Data <= 0.] = 0.
    data[np.isnan(data)] = 0.
    return data,hdr

def create_masks(data_in,hdr_in, working_dir, name, sofia_call='sofia2'):
    data = copy.deepcopy(data_in)
    hdr = copy.deepcopy(hdr_in)
    # First we smooth our template
    # We smooth this to 1.25 the input beam
    '''
    bmaj = hdr["BMAJ"]*3600.
    bmin = hdr["BMIN"]*3600.
    FWHM_conv_maj = np.sqrt((1.25 * bmaj) ** 2 - bmaj ** 2)
    FWHM_conv_min = np.sqrt((1.25 * bmin) ** 2 - bmin ** 2)
    # and in terms of pixels the sigmas
    sig_maj = (FWHM_conv_maj / np.sqrt(8 * np.log(2))) / \
               abs(hdr["CDELT1"] * 3600.)
    sig_min = (FWHM_conv_min / np.sqrt(8 * np.log(2))) / \
               abs(hdr["CDELT2"] * 3600.)
    '''
    #We replace zeros with NAN
    data[data == 0.] = float('NaN')
    # It seems that I do not smooth the mask anymore, why not?
    Tmp_Cube=data
    #Tmp_Cube = gaussian_filter(data, sigma=(0, sig_min, sig_maj), order=0)
    # Replace 0. with Nan

    #write this to the fits file
    fits.writeto(f'{working_dir}/tmp_{name}.fits',
                 Tmp_Cube, hdr, overwrite=True)
    SoFiA_Template = read_template_file('Sofia_Template.par')
    SoFiA_Template['input.data'.upper(
    )] = f'input.data = {working_dir}/tmp_{name}.fits'
    SoFiA_Template['scfind.threshold'.upper()] = 'scfind.threshold	= 3'
    SoFiA_Template['linker.minSizeZ'.upper(
    )] = f'linker.minSizeZ = {int(Tmp_Cube.shape[0]/2.)}'
    SoFiA_Template['output.filename'.upper()] = f'output.filename = tmp_{name}'
    with open(f'{working_dir}/tmp_{name}_sof.par', 'w') as file:
        file.writelines([SoFiA_Template[key] + "\n" for key in SoFiA_Template])
    run_sofia(working_dir, f'tmp_{name}_sof.par', sofia_call=sofia_call)
    #wait half a second to make sure got written
    #time.sleep(0.5)
    Mask_Outer = fits.open(f'{working_dir}/tmp_{name}_mask.fits')
    Mask_Outer[0].data[Tmp_Cube < 0.] = 0.
    #Mask_Outer[0].header['CUNIT3'] = 'KM/S'
    try:
        del Mask_Outer[0].header['HISTORY']
    except:
        pass
    #fits.writeto(f'{outdir}/Outer_{name}_mask.fits',Mask_Outer[0].data,Mask_Outer[0].header,overwrite = True)
    os.system(f"rm -f {working_dir}/tmp_{name}_mask.fits")
    # Create the inner mask
    SoFiA_Template['dilation.enable'.upper()] = 'dilation.enable	=	false'
    SoFiA_Template['scfind.threshold'.upper()] = 'scfind.threshold	= 5'
    with open(f'{working_dir}/tmp_{name}_sof.par', 'w') as file:
        file.writelines([SoFiA_Template[key] + "\n" for key in SoFiA_Template])
    run_sofia(working_dir, f'tmp_{name}_sof.par', sofia_call=sofia_call)
    #wait half a second to make sure got written
    time.sleep(0.5)
    Mask_Inner = fits.open(f'{working_dir}/tmp_{name}_mask.fits')
    #Mask_Inner[0].header['CUNIT3'] = 'KM/S'
    del Mask_Inner[0].header['HISTORY']
    #fits.writeto(f'{outdir}/Inner_{name}_mask.fits',Mask_Inner[0].data,Mask_Inner[0].header,overwrite = True)
    os.system(f"rm -f {working_dir}/tmp_{name}_mask.fits")
    os.system(f"rm -f {working_dir}/tmp_{name}.fits")
    os.system(f'rm -f {working_dir}/tmp_{name}_sof.par')
    transform = np.array(copy.deepcopy(Mask_Outer[0].data), dtype=float)
    transform[transform > 0.] = 1.
    tmp = gaussian_filter(transform, sigma=(0, 3, 3), order=0)
    transform = copy.deepcopy(tmp)
    tmp = []
    transform[Mask_Inner[0].data > 0.] = 1.
    transform = gaussian_filter(transform, sigma=(0, 1, 1), order=0)

    transform[transform < 0.1] = 0.
    del data
    del hdr
    Mask_Outer.close()
    Mask_Inner.close()

    return transform


create_masks.__doc__ = f'''
 NAME:
    create_masks

 PURPOSE:
    Create the masks for the ROC standards.

 CATEGORY:
    common_functions

 INPUTS:
    outdir = directory where to put the final masks for future use
    working_dir = directory where to run sofia
    name = Base name of the input and output files

 OPTIONAL INPUTS:
    sofia_call = command name to run sofia

 OUTPUTS:
    the cut cube is returned.

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''


def calculate_pixel_noise(requested_noise,smoothing_sigma,tolerance=0.025):
    #firt guess of the pixel noise
    pixel_noise = requested_noise *np.mean(smoothing_sigma)*2. * np.sqrt(np.pi)
    #  we need a shape for testing. something say 100 * smoothing kernel
    test_shape= np.array([int(x*100.) for x in smoothing_sigma],dtype=int)
    rng = np.random.default_rng()
    achieved_noise = 0.
    while abs(achieved_noise - requested_noise) / requested_noise > 0.025:
        #If not the first run update our pixel noise
        if achieved_noise != 0:
            pixel_noise = pixel_noise*requested_noise/achieved_noise
        #We only need to estimate this in a single channel there is no spectral component
        #fill with gaussian values
        test_noise = rng.normal(scale=pixel_noise, size=test_shape)
        # and smooth to the final beam
        test_noise_smoothed = gaussian_filter(test_noise,sigma=smoothing_sigma,\
                                order=0)
        achieved_noise = np.std(test_noise_smoothed)
        #print(f"The current pixel noise estimate leads to {achieved_noise} mJy/beam (Requested = {requested_noise} mJy/beam).")



    return achieved_noise,pixel_noise
calculate_pixel_noise.__doc__ = f'''
 NAME:
    calculate_pixel_noise

 PURPOSE:
    calculate the noise that is required as input to random value generator to end up with the required noise after smoothing

 CATEGORY:
    common_functions

 INPUTS:
    requested_noise = the finale requested noise in smoothed images
    smoothing_sigma = the sigma of the final beam in pixels


 OPTIONAL INPUTS:
    tolerance = the accepted difference between the requested noise and the achieved noise

 OUTPUTS:
    achieved_noise = the smoothed standard deviation
    pixel_noise = the input pixel standard deviation

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''



def cut_input_cube(file_in, sizes, debug=False):
    cube = fits.open(file_in, uint=False,
                     do_not_scale_image_data=True, ignore_blank=True)
    hdr = cube[0].header
    for i, pos in enumerate(sizes):
        if pos[1] == -1:
            sizes[i][1] = hdr[f'NAXIS{3-i}']-1

    if hdr['NAXIS'] > 3:
        hdr['NAXIS'] = 3
        del hdr['*4']
        new_data = cube[0].data[0, sizes[0][0]:sizes[0][1],
            sizes[1][0]:sizes[1][1], sizes[2][0]:sizes[2][1]]
    else:
        new_data = cube[0].data[sizes[0][0]:sizes[0][1],
            sizes[1][0]:sizes[1][1], sizes[2][0]:sizes[2][1]]

    if f'PC01_01' in hdr:
        del hdr['PC0*_0*']
    hdr['NAXIS1'] = sizes[2][1]-sizes[2][0]
    hdr['NAXIS2'] = sizes[1][1]-sizes[1][0]
    hdr['NAXIS3'] = sizes[0][1]-sizes[0][0]

    hdr['CRPIX1'] = hdr['CRPIX1']-sizes[2][0]
    hdr['CRPIX2'] = hdr['CRPIX2']-sizes[1][0]
    hdr['CRPIX3'] = hdr['CRPIX3']-sizes[0][0]
    try:
        del hdr['HISTORY']
    except:
        pass
    if hdr['BITPIX'] < -32:
        hdr['BITPIX'] = -32
        #print(hdr['BITPIX'])
        #exit()
    elif hdr['BITPIX'] > -32:
        hdr['BITPIX'] = 32
    # Do not regid the original cubes if we do this we do this then we mess up the barollo and rc files
    cube[0].header=hdr
    cube[0].data=np.array(new_data, dtype = np.float32)
    return cube
cut_input_cube.__doc__=f'''
 NAME:
    cut_input_cube

 PURPOSE:
    Cut filename back to the size of subcube, update the header and write back to disk.

 CATEGORY:
    common_functions

 INPUTS:
    file_in = the fits file to cut
    sizes = array that contains the new size as
                [[z_min,z_max],[y_min,y_max], [x_min,x_max]]
                adhering to fits' idiotic way of reading fits files.


 OPTIONAL INPUTS:
    debug = False

 OUTPUTS:
    the cut cube is returned.

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''

def delete_directory(dir_to_delete):
    for f in os.listdir(dir_to_delete):
        try:
            os.remove(os.path.join(dir_to_delete, f))
        except FileNotFoundError:
            pass
        except IsADirectoryError:
            delete_directory(os.path.join(dir_to_delete, f))

    try:
        os.rmdir(dir_to_delete)
    except FileNotFoundError:
        pass
delete_directory.__doc__=f'''
 NAME:
    delete_directory

 PURPOSE:
    delete a non empty directory

 CATEGORY:
    common_functions

 INPUTS:
    dir = directory to delete

 OPTIONAL INPUTS:

 OUTPUTS:

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''


def download_cube(name, url, sizes, new_location):
    name_in=download_file(url, pkgname = 'pyHIARD', timeout = 1800)
    cube=cut_input_cube(name_in, sizes, name = name)
    fits.writeto(f'{new_location}/{name}.fits', cube[0].data, cube[0].header)
    clear_download_cache(url, pkgname = 'pyHIARD')
    return cube
download_cube.__doc__=f'''
 NAME:
    download_cube

 PURPOSE:
    Download the specified cube when not present in installation and cut down required size
    Store in installation for future use.

 CATEGORY:
    common_functions

 INPUTS:
    name = local name of the cube
    url = url of location of remote cube
    sizes = require sizes for stored cube
    new_location = location of correct directory in installation.


 OPTIONAL INPUTS:
    debug = False

 OUTPUTS:
    the cut cube is returned

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''

def find_program(name, search):
    found=False
    while not found:
        try:
            run=subprocess.Popen(
                [name], stdout = subprocess.PIPE, stderr = subprocess.PIPE)
            run.stdout.close()
            run.stderr.close()
            os.kill(run.pid, signal.SIGKILL)
            found=True
        except Exception as e:
            traceback.print_tb(e.__traceback__)
            name=input(f'''You have indicated to use {name} for using {search} but it cannot be found.
Please provide the correct name : ''')
    return name
find_program.__doc__=f'''
 NAME:
    find_program

 PURPOSE:
    check whether a program is available for use.

 CATEGORY:
    support_functions

 INPUTS:
    name = command name of the program to run
    search = Program we are looking for
 OPTIONAL INPUTS:

 OUTPUTS:
    the correct command for running the program

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''


def get_beam_area_in_pixels(Template_Header, beam= None):

    if beam is None:
        beam = [Template_Header["BMAJ"],Template_Header["BMIN"]]
    #  https://science.nrao.edu/facilities/vla/proposing/TBconv
    beamarea=(np.pi*abs((beam[0]*beam[1])))/(4.*np.log(2.))
    return beamarea/(abs(Template_Header['CDELT1'])*abs(Template_Header['CDELT2']))
get_beam_area_in_pixels.__doc__=f'''
NAME:
    get_beam_in_pixels(Template_Header, beam= [-1,-1.])

PURPOSE:
    calculate the area of the beam in pixels

CATEGORY:
    common_functions

INPUTS:
    name = basename of template
    beams = the requested beam in degrees, if unset it assumed that the beam is in the header.

OPTIONAL INPUTS:

OUTPUTS:
    the beam area in pixels

OPTIONAL OUTPUTS:

PROCEDURES CALLED:
   Unspecified

NOTE:
'''
# Function to input a boolean answer
def get_bool(print_str = "Please type True or False", default = True):
    invalid_input=True
    while invalid_input:
        inp=input(print_str)
        if inp == "":
            if default:
                return True
            else:
                return False
        elif inp.lower() == "true" or inp.lower() == "t" or inp.lower() == "y" or inp.lower() == "yes":
            return True
        elif inp.lower() == "false" or inp.lower() == "f" or inp.lower() == "n" or inp.lower() == "no":
            return False
        else:
            print("Error: the answer must be true/false or yes/no.")
get_bool.__doc__=f'''
NAME:
    get_bool

PURPOSE:
    get a versatile boolean response from the user

CATEGORY:
    common_functions

INPUTS:

OPTIONAL INPUTS:
    print_str = "Please type True or False"
    The question to ask the user

    default = True
    The default answer will correspond to the default

OUTPUTS:
    The corresponding boolean

OPTIONAL OUTPUTS:

PROCEDURES CALLED:
Unspecified

NOTE:
'''

def get_created_models(Catalogue, delete_existing):
    cat_exists=os.path.isfile(Catalogue)
    if not delete_existing and cat_exists:
        with open(Catalogue) as cat:
            lines=cat.readlines()
        try:
            split_line=lines[-1].split('|')
            if len(split_line) > 0:
                number_models=int(float(split_line[0])+1)
        except:
            number_models=0
    else:
        with open(Catalogue, 'w') as cat:
            cat.write('ID|Distance|Directoryname|Cubename\n')
        number_models=0
    return number_models
get_created_models.__doc__=f'''
 NAME:
    get_created_models(Catalogue,delete_exiting)

 PURPOSE:
    check whether a new catalogue should be created if not then the starting number is the end of the catalogue

 CATEGORY:
    support_functions

 INPUTS:
    name = command name of the program to run
    search = Program we are looking for
 OPTIONAL INPUTS:

 OUTPUTS:
    the correct command for running the program

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''

def get_mask(Cube_In, factor = 5.2):
    Mask = copy.deepcopy(Cube_In)
    Top_Cube = Cube_In[Cube_In > 0.9*np.max(Cube_In)]
    Sorted_Cube = Top_Cube[np.argsort(Top_Cube)]
    Max_Flux = np.mean(Sorted_Cube[-20:])
    Mask[Cube_In > 0.005*Max_Flux] = 1.
    Mask[Cube_In < 0.005*Max_Flux] = 0.
    # If we have a lot of pixels and stuff we need smooth more
    #So we take the inverse of the initial factor with a standard of 0.75 minimum of 0.5 and a maximum of 1.25
    pix_smooth = 3.9/factor
    if pix_smooth < 0.5:
        pix_smooth=0.5
    if pix_smooth > 3:
        pix_smooth=3.
    Mask =  gaussian_filter(Mask,sigma=(0.,pix_smooth,pix_smooth),order=0)
    Mask[Mask > 0.95] = 1
    Mask[Mask < 0.05] = 0
    Mask = np.array(Mask,dtype=int)
    return Mask

def get_mean_flux(Cube_In,Mask = None):

    #As python is really the dumbest language ever invented there is different behaviour for passing np.arrays and lists
    Cube=copy.deepcopy(Cube_In)
    # To calculate the mean of all values in the mask is too sensitive to very small variations in the mask
    if Mask is not None:
        Cube[Mask < 0.5] = 0.
    # First we need to get a maximum flux value.
    Top_Cube = Cube[Cube > 0.9*np.max(Cube)]
    Sorted_Cube = Top_Cube[np.argsort(Top_Cube)]
    Max_Flux = np.mean(Sorted_Cube[-20:])
    Mean = np.mean(Cube[Cube > 0.07*Max_Flux])
    # Regridded Our new method gives 0.03289582223850146. The old gave 0.026361946919571028
    #Our new method gives 0.03307503935880944. The old gave 0.025313454549815356
    return Mean
get_mean_flux.__doc__=f'''
 NAME:
    get_created_models(Catalogue,delete_exiting)

 PURPOSE:
    check whether a new catalogue should be created if not then the starting number is the end of the catalogue

 CATEGORY:
    support_functions

 INPUTS:
    name = command name of the program to run
    search = Program we are looking for
 OPTIONAL INPUTS:

 OUTPUTS:
    the correct command for running the program

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''
def limit_memory(maxsize):
    soft, hard=resource.getrlimit(resource.RLIMIT_AS)
    resource.setrlimit(resource.RLIMIT_AS, (maxsize, hard))

def load_text_model(filename, type = 'Tirific', Variables = None ,\
                        package_file = True):
    if Variables is None:
        Variables = ['BMIN', 'BMAJ', 'BPA', 'RMS', 'DISTANCE', 'NUR', 'RADI', \
            'VROT','Z0', 'SBR', 'INCL', 'PA', 'XPOS', 'YPOS', 'VSYS', 'SDIS', \
            'VROT_2',  'Z0_2', 'SBR_2','INCL_2', 'PA_2', 'XPOS_2', 'YPOS_2', \
            'VSYS_2', 'SDIS_2', 'CONDISP', 'CFLUX', 'CFLUX_2']

    #First check that we have a proper type
    allowed_types=['tir', 'rc', 'bar', 'tirific', 'rotcur', 'barolo', 'fat']
    while type.lower() not in allowed_types:
        type=input(f'''pyHIARD can not deal with the type {type}.
please provide one of the following types {', '.join(allowed_types)}:''')
    # Then make sure the type adheres to a single desganation
    if type.lower() in ['tir', 'tirific']:
        type='tir'
    elif type.lower() in ['rc', 'rotcur']:
        type='rc'
    elif type.lower() in ['bar', 'barolo']:
        type='bar'
    else:
        raise InputError('This should be impossible.')

    ext={'tir': 'def', 'bar': 'txt', 'rc': 'rotcur'}
    if package_file:
        python_version = sys.version_info
        if python_version[0] < 3. or \
            (python_version[0] == 3. and python_version[1] < 9.):
            model=__import__(
                f'pyHIARD.Resources.Cubes.{filename}', globals(), locals(), filename, 0)
            with import_res.open_text(model, f'{filename}.{ext[type]}') as tmp:
                unarranged = tmp.readlines()
        else:

            from importlib import import_module
            model = import_module( f'pyHIARD.Resources.Cubes.{filename}')
            with import_res.files(model).joinpath(f'{filename}.{ext[type]}').open('r') as tmp:
                unarranged = tmp.readlines()

    else:
        with open(filename, 'r') as tmp:
            unarranged=tmp.readlines()
    Values=Proper_Dictionary({})
    variable_location={}

    for key in Variables:
        Values[key]=[]
        variable_location[key]=-1
    mapping={'rc': {'RADI': 'radius', 'VSYS': 'systemic', 'VSYS_ERR': 'systemic_error',
                      'VROT': 'rotation', 'VROT_ERR': 'rotation_error', 'VRAD': 'expansion',
                      'VRAD_ERR': 'expansion_error', 'PA': 'pos.', 'PA_ERR': 'pos._error',
                      'INCL': 'incli-', 'INCL_ERR': 'incli-_error', 'XPOS': 'x-pos.',
                      'XPOS_ERR': 'x-pos._error', 'YPOS': 'y-pos.', 'YPOS_ERR': 'y-pos._error'},
               'bar': {'RADI': 'RAD(arcs)', 'VSYS': 'VSYS(km/s)', 'VSYS_ERR': 'E_VSYS',
                      'VROT': 'VROT(km/s)', 'VROT_ERR': 'E_VROT', 'VRAD': 'VRAD(km/s)',
                      'VRAD_ERR': 'E_VRAD', 'PA': 'P.A.(deg)', 'PA_ERR': 'E_PA',
                      'INCL': 'INC(deg)', 'INCL_ERR': 'E_INC', 'XPOS': 'XPOS(pix)',
                      'XPOS_ERR': 'E_XPOS', 'YPOS': 'YPOS(pix)', 'YPOS_ERR': 'E_YPOS',
                      'SDIS': 'DISP(km/s)', 'SDIS_ERR': 'E_DISP', 'Z0': 'Z0(arcs)', 'Z0_ERR': 'E_Z0'}
                      }
    for line in unarranged:
        stripped = False
        if line[0] in ['#', '!']:
            line = line[1:]
            stripped = True
        if type == 'tir':
            variable = str(line.split('=')[0].strip().upper())
            if variable in Variables:
                Values[variable] = [float(x)
                                          for x in line.split('=')[1].rsplit()]
        elif stripped:
            split_line= [x.strip() for x in line.split()]
            if len(split_line) == 0:
                continue
            if type == 'rc':
                if split_line[0] != 'radius':
                    continue
            for key in Variables:
                if key not in mapping[type]:
                    continue
                if key[-4: ] == '_ERR':
                    if type == 'bar':
                        if f'{mapping[type][key]}1' in split_line:
                            variable_location[key]= [split_line.index(f'{mapping[type][key]}1'), split_line.index(f'{mapping[type][key]}2')]
                    elif type == 'rc':
                        var= mapping[type][key].split('_')
                        if var[0] in split_line:
                            if split_line[split_line.index(var[0])+1] == var[1]:
                                variable_location[key]= split_line.index(
                                    var[0])+1
                else:
                    if mapping[type][key] in split_line:
                        variable_location[key] = split_line.index(
                            mapping[type][key])
        else:
            split_line = [float(x.strip()) for x in line.split()]
            for var in variable_location:
                if variable_location[var] == -1:
                    pass
                elif isinstance(variable_location[var], int):
                    Values[var].append(
                        float(split_line[variable_location[var]]))
                else:
                    the_av = np.mean([abs(split_line[variable_location[var][0]]), abs(split_line[variable_location[var][1]])])
                    Values[var].append(float(the_av))

    final_values = []
    for key in Variables:
        if len(Values[key]) > 0.:
            if 'RADI' in Variables:
                while len(Values[key]) < len(Values['RADI']):
                    Values[key].append(Values[key][-1])
            final_values.append(np.array(Values[key]))
        else:
            if 'RADI' in Variables:
                final_values.append(np.full(len(Values['RADI']), -1.))
            else:
                final_values.append(np.full(len(Values[0]), -1.))
    return final_values
load_text_model.__doc__ =f'''
 NAME:
    load_text_model

 PURPOSE:
    load a text file tilted ring model

 CATEGORY:
    common_functions

 INPUTS:
    name = name of the packaged galaxy or text file
    type = type of input file, options are  ['tir','rc','bar','tirific','rotcur','barolo','fat']
    Variables = variable to extract following tirific convention
    package_file = Boolean indicating whether the file is from the python package or not

 OPTIONAL INPUTS:

 OUTPUTS:
    arrays of variables in the order of variables is put in.

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
    Note that rotcur and barolo are not very intelligent and hence return the
    xpos and ypos in pixel coordinates which are worthless without a corresponding cube.
    barolo errors are averaged from lower and upper error in the file.

 '''
#

def plot_input(directory, Model,add_sbr = None, Distance= 0., RHI = None \
                ,Title = 'EMPTY',WarpR=None, font_file = 'empty.ttf'):

    if add_sbr is None:
        add_sbr = [0.,0.]
    if RHI is None:
        RHI = [0.,0.,0.]
    if WarpR is None:
        WarpR=[0.,0.]
    variables_to_plot = ['SBR', 'VROT','PA','INCL','SDIS','Z0']
    plots = len(variables_to_plot)
    units = {'SBR': 'SBR (Jy km s$^{-1}$ arcsec$^{-2}$)' ,
            'VROT': 'V$_{rot}$ (km s$^{-1}$)',
            'PA': 'PA ($^{\circ}$)',
            'INCL': 'INCL ($^{\circ}$)',
            'SDIS': 'Disp. (km s$^{-1}$)',
            'Z0': 'Z0 (arcsec)'}
    plt.figure(2, figsize=(8, 12), dpi=100, facecolor='w', edgecolor='k')
    try:
        mpl_fm.fontManager.addfont(font_file)
        font_name = mpl_fm.FontProperties(fname=font_file).get_name()
    except FileNotFoundError:
        font_name = 'DejaVu Sans'
    labelfont = {'family': font_name,
                'weight': 'normal',
                'size': 18}
    plt.rc('font', **labelfont)
    radius = np.array([float(x) for x in Model['RADI'].split('=')[1].split()], dtype=float)
    bmaj = float(Model['BMAJ'].split('=')[1])

    radius_bmaj = [0]
    counter = 1.
    for i, rad in enumerate(radius):
        if rad < counter*bmaj:
            pass
        else:
            radius_bmaj.append(i)
            counter += 1.
    radius_bmaj = np.array(radius_bmaj,dtype=int)
    rad_unit = 'arcsec'
    #if we have a distance we convert all arcsec values to kpc, It is assumed all extra variables come in the right unit
    if Distance > 0.:
        radius = convertskyangle(radius, distance=Distance)
        rad_unit = 'kpc'
        units['Z0'] = 'Z0 (kpc)'
    for i, variable in enumerate(variables_to_plot):
        plt.subplot(plots, 1, plots-i)
        var_to_plot = [float(x) for x in Model[variable].split('=')[1].split()]
        while len(var_to_plot) < len(radius):
            var_to_plot.append(var_to_plot[-1])
        var_to_plot = np.array(var_to_plot,dtype=float)
        if variable == 'Z0' and Distance != 0.:
            var_to_plot = convertskyangle(var_to_plot , distance=Distance)
        plt.plot(radius, var_to_plot, 'k')
        plt.plot(radius[radius_bmaj], var_to_plot[radius_bmaj], 'ko')
        var_to_plot2 = [float(x) for x in Model[f'{variable}_2'].split('=')[1].split()]
        while len(var_to_plot2) < len(radius):
            var_to_plot2.append(var_to_plot2[-1])
        if variable == 'Z0' and Distance != 0.:
                var_to_plot2 = convertskyangle(var_to_plot2 , distance=Distance)
        if np.sum(var_to_plot2) > 0. and np.sum([x-y for x, y in zip(var_to_plot,var_to_plot2)]) != 0.:
            var_to_plot = np.array(var_to_plot2,dtype=float)
            plt.plot(radius, var_to_plot, 'r')
            plt.plot(radius[radius_bmaj], var_to_plot[radius_bmaj], 'ro')
        if i == 0.:
            plt.xlabel(f'Radius ({rad_unit})', **labelfont)
            lab_bottom = True
        else:
            lab_bottom = False
        if variable in ['SBR'] and np.sum(add_sbr) != 0:
            plt.plot(radius, add_sbr[0, :], 'b')
            plt.plot(radius[radius_bmaj], add_sbr[0, radius_bmaj], 'bo')
            plt.plot(radius, add_sbr[1, :], 'y')
            plt.plot(radius[radius_bmaj], add_sbr[1, radius_bmaj], 'yo')

        plt.ylabel(units[variable], **labelfont)
        ymin,ymax = plt.ylim()
        plt.margins(x=0., y=0.)
        if variable in ['SBR', 'VROT','SDIS','Z0'] and RHI[0] != 0:
            plt.plot([RHI[0], RHI[0]],[ymin-(ymax-ymin)*0.1,ymax+(ymax-ymin)*0.1],'b')
            if variable in ['SBR'] and RHI[1] != 0:
                plt.plot([RHI[1], RHI[1]], [ymin - (ymax - ymin)
                         * 0.1, ymax + (ymax - ymin) * 0.1], 'b--')
            if variable in ['SBR'] and RHI[2] != 0:
                plt.plot([RHI[2], RHI[2]], [ymin - (ymax - ymin)
                         * 0.1, ymax + (ymax - ymin) * 0.1], 'b--')
        if variable in ['INCL', 'PA'] and np.sum(WarpR) != 0:
            plt.plot([WarpR[1], WarpR[1]],[ymin-(ymax-ymin)*0.1,ymax+(ymax-ymin)*0.1],'g')
            if WarpR[0] != 0.:
                plt.plot([WarpR[0], WarpR[0]],[ymin-(ymax-ymin)*0.1,ymax+(ymax-ymin)*0.1],'g--')
        plt.tick_params(
        axis = 'x',  # changes apply to the x-axis
        which = 'both',  # both major and minor ticks are affected
        direction = 'in',
        bottom = True,  # ticks along the bottom edge are off
        top = False,  # ticks along the top edge are off
        labelbottom =lab_bottom)  # labels along the bottom edge are off
    if Title != 'EMPTY':
        plt.title(Title)
    plt.savefig(f"{directory}Overview_Input.png", bbox_inches ='tight')
    plt.close()
plot_input.__doc__= f'''
NAME:
    plot_input(directory,Model,add_sbr = 0, Distance= 0.,
               RHI = [0.,0.,0.] ,Title = 'EMPTY', font_file='empty.ttf')

PURPOSE:
    Plot the tirific template as output

CATEGORY:
   roc

INPUTS:
    directory = the destination of the plot
    Model = Tirific Template
    add_sbr = additional SBR profiles to plot
    Distance = Distance to have kpc instead of arcsec
    RHI = The radius corresponding to RHI [symmetric,appr,rec]
    Warp_Start = radius corresponding to the start of the warp.
    Title = title string
    font_file = location of the true type font to be used for plotting.

OPTIONAL INPUTS:

OUTPUTS:
    standardized plot for output

OPTIONAL OUTPUTS:

PROCEDURES CALLED:
   Unspecified

NOTE:
'''
def print_base_galaxy(Galaxy):
    try:
        RAhr, DEChr= convertRADEC(Galaxy.Coord[0], Galaxy.Coord[1])
    except AttributeError:
        RAhr, DEChr= 'Random', 'Random'
    print(f'''The galaxy has the central coordinates RA= {RAhr}, DEC={DEChr}
{'Inclination':15s} = {Galaxy.Inclination:<10.1f}, {'Dispersion':15s} = {Galaxy.Dispersion}
{'Mass':15s} = {Galaxy.Mass:<10.2e}, {'PA':15s} = {Galaxy.PA}
{'Beams':15s} = {Galaxy.Beams:<10.1f}, {'Warp':15s} = {Galaxy.Warp}
{'Radial_Motions':15s} = {Galaxy.Radial_Motions:<10.1f}, {'SNR':15s} = {Galaxy.SNR}
{'Channel Width':15s} = {Galaxy.Channelwidth:<10.1f}, {'Beam Size':15s} = {Galaxy.Res_Beam}
{'Arms':15s} = {Galaxy.Arms:10s}, {'Flare':15s} = {Galaxy.Flare:15s}, {'Bar':15s} = {Galaxy.Bar:10s}
''')


print_base_galaxy.__doc__=f''' NAME:
    print_base_galaxy

 PURPOSE:
    print the contents of a galaxy class in an orderly fashion

 CATEGORY:
    common_functions

 INPUTS:
    galaxy class

 OPTIONAL INPUTS:

 OUTPUTS:
    print out of the galaxies parameters.

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
 '''

def scrambled_initial(directory, Model):
    variables=['VROT', 'SBR','VRAD','SDIS','PA','INCL','XPOS','YPOS','VSYS','Z0']
    profile= {}

    for var in variables:
        if var in Model:
            ini= [float(x) for x in Model[var].split('=')[1].split()]
            ini2=[float(x) for x in Model[f'{var}_2'].split('=')[1].split()]
            while len(ini) < len(ini2):
                ini.append(ini[-1])
            while len(ini) > len(ini2):
                ini2.append(ini2[-1])
            if var == 'VROT':
                profile[var]=np.mean([ini[-5:-1], ini2[-5:-1]])
            elif var == 'SBR':
                profile[var] = np.max([ini, ini2])
            else:
                profile[var]=np.mean([ini[0], ini2[0]])
        else:
            profile[var]= 0.
    rng = np.random.default_rng()
    svrot= profile['VROT']+rng.uniform(-10., 10.)
    sincl= profile['INCL']+rng.uniform(-10., 10.)
    spa=  profile['PA']+rng.uniform(-3., 3.)
    sz0= profile['Z0']
    ssbr= rng.uniform(-1*profile['SBR'], profile['SBR'])
    ssdis= profile['SDIS']+rng.uniform(-4., 4.)
    svrad=profile['VRAD']+rng.uniform(-10., 10.)
    sra=profile['XPOS']+rng.uniform(-10./3600., -10./3600.)
    sdec=profile['YPOS']+rng.uniform(-10./3600., -10./3600.)
    svsys= profile['VSYS']+rng.uniform(-4., 4.)
    with open(f"{directory}Initial_Estimates.txt", 'w') as overview:
        overview.write(f'''#This file contains the random varied initial estimates.
#{'VROT':<15s} {'INCL':<15s} {'PA':<15s} {'Z0':<15s} {'SBR':<15s} {'DISP':<15s} {'VRAD':<15s} {'RA':<15s} {'DEC':<15s}  {'VSYS':<15s}
#{'km/s':<15s} {'Degree':<15s} {'Degree':<15s} {'arcsec':<15s} {'Jy km/s/arcsec^2':<15s} {'km/s':<15s} {'km/s':<15s} {'Degree':<15s} {'Degree':<15s}  {'km/s':<15s}
{svrot:<15.2f} {sincl:<15.2f} {spa:<15.2f} {sz0:<15.3f} {ssbr:<15.7f} {ssdis:<15.2f} {svrad:<15.2f} {sra:<15.5f} {sdec:<15.5f}  {svsys:<15.2f}''')

scrambled_initial.__doc__=f'''
NAME:
   scrambled_initial(directory,Model)

PURPOSE:
    Create a file with scrambelled Initial estimates.

CATEGORY:
   common_functions

INPUTS:
    directory = the directory where to put the model
    Model = array with def template with all lines split
OPTIONAL INPUTS:

OUTPUTS:
    text file

OPTIONAL OUTPUTS:

PROCEDURES CALLED:
   Unspecified

NOTE:
'''

#Function to read simple input files that  use = as a separator between ithe required input and the values

def read_template_file(filename, package_file = True):
    if float(sys.version[:3]) < 3.9 and package_file:
        with import_res.open_text(templates, filename) as tmp:
            unarranged = tmp.readlines()
    elif package_file:
        with import_res.files(templates).joinpath(filename).open('r') as tmp:
            unarranged = tmp.readlines()
    else:
        with open(filename, 'r') as tmp:
            unarranged= tmp.readlines()
    Template_in= Proper_Dictionary({})

    # Separate the keyword names
    for tmp in unarranged:
        # python is really annoying with needing endlines. Let's strip them here and add them when writing
        Template_in[tmp.split('=', 1)[0].strip().upper()]=tmp.rstrip()
    return Template_in

def rotateCube(Cube, angle, pivot,order=1):
    padX= [int(Cube.shape[2] - pivot[0]), int(pivot[0])]
    padY= [int(Cube.shape[1] - pivot[1]), int(pivot[1])]
    imgP= np.pad(Cube, [[0, 0], padY, padX], 'constant')
    #Use nearest neighbour as it is exact enough and doesn't mess up the 0. and is a lot faster
    imgR = rotate(imgP, angle, axes =(2, 1), reshape=False,order=order)
    return imgR[:, padY[0]: -padY[1], padX[0]: -padX[1]]
rotateCube.__doc__=f'''
 NAME:
    rotateCube(Cube, angle, pivot)

 PURPOSE:
    rotate a cube in the image plane

 CATEGORY:
    common_functions

 INPUTS:
    Cube = the cube data array
    angle = the angle to rotate under
    pivot = the point around which to rotate

 OPTIONAL INPUTS:

 OUTPUTS:
    the rotated cube

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''
def regrid_array(oldarray, Out_Shape):
    oldshape = np.array(oldarray.shape)
    newshape = np.array(Out_Shape, dtype=float)
    ratios = oldshape/newshape
        # calculate new dims
    nslices = [ slice(0,j) for j in list(newshape) ]
    #make a list with new coord
    new_coordinates = np.mgrid[nslices]
    #scale the new coordinates
    for i in range(len(ratios)):
        new_coordinates[i] *= ratios[i]
    #create our regridded array
    newarray = map_coordinates(oldarray, new_coordinates,order=1)
    if any([x != y for x,y in zip(newarray.shape,newshape)]):
        print("Something went wrong when regridding.")
    return newarray
regrid_array.__doc__ =f'''
 NAME:
regridder
 PURPOSE:
Regrid an array into a new shape through the ndimage module
 CATEGORY:
    fits_functions

 INPUTS:
    oldarray = the larger array
    newshape = the new shape that is requested

 OPTIONAL INPUTS:

 OUTPUTS:
    newarray = regridded array

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    scipy.ndimage.map_coordinates, np.array, np.mgrid

 NOTE:
'''

def run_sofia(working_dir, parameter_file,sofia_call = 'sofia2'):
    sfrun = subprocess.Popen([sofia_call,parameter_file],cwd =working_dir, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
    sofia_run, sofia_warnings_are_annoying= sfrun.communicate()
    print(sofia_run.decode("utf-8"))
    if sfrun.returncode == 8:
        raise SofiaFaintError(
            "RUN_SOFIA:Sofia cannot find a source in the input cube")
    elif sfrun.returncode == 0:
        sofia_ok= True
    else:
        print(sofia_warnings_are_annoying.decode("utf-8"))
        raise SofiaRunError(
            "RUN_SOFIA:Sofia did not execute properly. See screen for details")
run_sofia.__doc__=f'''
 NAME:
    run_sofia

 PURPOSE:
    run sofia

 CATEGORY:
    common_functions

 INPUTS:
    working_dir = directory where to run sofia
    parameter_file = input parameters file for sofia


 OPTIONAL INPUTS:
    sofia_call = command name for running sofia

 OUTPUTS:

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''
